#
# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for RecipeBase."""

from ..metarecipes import RecipeType
from ..recipes import BaseRecipe
from ..recipeinout import RecipeInput, RecipeResult
from ..requirements import ObservationResultRequirement
from ..dataholders import Result
from ..requirements import Requirement
from numina.types.qc import QC
from numina.types.datatype import PlainPythonType


class PruebaRecipe1(BaseRecipe):
    somereq = Requirement(int, 'Some integer')
    someresult = Result(int, 'Some integer')

    def run(self, recipe_input):
        if recipe_input.somereq >= 100:
            result = 1
        else:
            result = 100

        return self.create_result(someresult=result)

    def run_qc(self, recipe_input, recipe_result):
        if recipe_result.someresult >= 100:
            recipe_result.qc = QC.BAD
        else:
            recipe_result.qc = QC.GOOD

        return recipe_result


def test_metaclass_empty_base():

    class RecipeTest(metaclass=RecipeType):
        pass

    assert hasattr(RecipeTest, 'RecipeInput')

    assert hasattr(RecipeTest, 'RecipeResult')

    assert issubclass(RecipeTest.RecipeInput, RecipeInput)

    assert issubclass(RecipeTest.RecipeResult, RecipeResult)

    assert RecipeTest.RecipeInput.__name__ == 'RecipeInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeResult'


def test_metaclass():

    class RecipeTest(metaclass=RecipeType):
        obsresult = ObservationResultRequirement()
        someresult = Result(int, 'Some integer')

    assert hasattr(RecipeTest, 'RecipeInput')

    assert hasattr(RecipeTest, 'RecipeResult')

    assert RecipeTest.RecipeInput.__name__ == 'RecipeTestInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeTestResult'


def test_recipe_with_autofield():

    class RecipeTestAutoField(BaseRecipe):
        qc82h = Result(float, destination='qc')

    class RecipeTest(RecipeTestAutoField):
        obsresult = ObservationResultRequirement()
        someresult = Result(int, 'Some integer')

    assert hasattr(RecipeTest, 'RecipeInput')

    assert hasattr(RecipeTest, 'RecipeResult')

    assert issubclass(RecipeTest.RecipeInput, RecipeInput)

    assert issubclass(RecipeTest.RecipeResult, RecipeResult)

    assert RecipeTest.RecipeInput.__name__ == 'RecipeTestInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeTestResult'

    assert 'qc' in RecipeTest.RecipeResult.stored()
    assert 'qc' in RecipeTest.products()

    for prod in RecipeTest.RecipeResult.stored().values():
        assert isinstance(prod, Result)

    qc = RecipeTest.RecipeResult.stored()['qc']

    assert isinstance(qc.type, PlainPythonType)


def test_recipe_without_autofield():

    class RecipeTest(BaseRecipe):
        obsresult = ObservationResultRequirement()
        someresult = Result(int, 'Some integer')

    assert hasattr(RecipeTest, 'RecipeInput')

    assert hasattr(RecipeTest, 'RecipeResult')

    assert issubclass(RecipeTest.RecipeInput, RecipeInput)

    assert issubclass(RecipeTest.RecipeResult, RecipeResult)

    assert RecipeTest.RecipeInput.__name__ == 'RecipeTestInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeTestResult'

    assert 'qc' not in RecipeTest.RecipeResult.stored()
    assert 'qc' not in RecipeTest.products()

    for prod in RecipeTest.RecipeResult.stored().values():
        assert isinstance(prod, Result)


def test_recipe_io_inheritance():

    class TestBaseRecipe(BaseRecipe):
        obresult = ObservationResultRequirement()
        someresult1 = Result(int, 'Some integer')

    class RecipeTest(TestBaseRecipe):
        other = Requirement(int, description='Other')
        someresult2 = Result(int, 'Some integer')

    assert issubclass(RecipeTest.RecipeInput, TestBaseRecipe.RecipeInput)

    assert issubclass(RecipeTest.RecipeResult, TestBaseRecipe.RecipeResult)

    assert RecipeTest.RecipeInput.__name__ == 'RecipeTestInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeTestResult'

    assert 'obresult' in RecipeTest.requirements()
    assert 'other' in RecipeTest.requirements()
    assert 'someresult1' in RecipeTest.products()
    assert 'someresult2' in RecipeTest.products()


def test_recipe_io_baseclass():

    class MyRecipeInput(RecipeInput):
        def myfunction(self):
            return 1

    class MyRecipeResult(RecipeResult):
        def myfunction(self):
            return 2

    class RecipeTest(BaseRecipe):

        RecipeInput = MyRecipeInput

        RecipeResult = MyRecipeResult

        other = Requirement(int, description='Other')
        someresult2 = Result(int, 'Some integer')

    assert issubclass(RecipeTest.RecipeInput, MyRecipeInput)

    assert issubclass(RecipeTest.RecipeResult, MyRecipeResult)

    assert RecipeTest.RecipeInput.__name__ == 'RecipeTestInput'

    assert RecipeTest.RecipeResult.__name__ == 'RecipeTestResult'

    assert 'other' in RecipeTest.requirements()
    assert 'someresult2' in RecipeTest.products()

    assert RecipeTest.RecipeInput().myfunction() == 1
    assert RecipeTest.RecipeResult().myfunction() == 2


def test_run_qc():

    recipe_input = PruebaRecipe1.create_input(somereq=100)
    recipe = PruebaRecipe1()
    result = recipe(recipe_input)

    assert result.qc == QC.GOOD


def test_run_base():

    recipe_input = PruebaRecipe1.create_input(somereq=1)
    recipe = PruebaRecipe1()
    result = recipe(recipe_input)

    assert result.qc == QC.BAD
    assert result.someresult == 100



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute image3d IFU from RSS image"""
import numpy as np
from scipy.signal import convolve2d

from numina.tools.ctext import ctext


def compute_image3d_ifu_from_rss_method1(
    image2d_rss_method1,
    naxis1_detector,
    naxis2_ifu,
    naxis1_ifu,
    nslices,
    verbose=False   
):
    """
    Compute the image3d IFU from the image2d RSS using method 1.
    
    Parameters
    ----------
    image2d_rss_method1 : numpy.ndarray
        The input RSS image.
    naxis1_detector : int
        Number of pixels in the first axis of the detector image
        (dispersion direction).
    naxis2_ifu : int
        Number of pixels in the second axis of the IFU image.
    naxis1_ifu : int
        Number of pixels in the first axis of the IFU image.
    nslices : int
        Number of slices in the IFU image.
    verbose : bool, optional
        If True, print verbose output. Default is False.
    
    Returns
    -------
    image3d_ifu_method1 : numpy.ndarray
        The output IFU image.
    """

    if verbose:
        print(ctext('\n* Computing image3d IFU from image2d RSS method 1', fg='green'))

    # kernel in the spectral direction
    # (bidimensional to be applied to a bidimensional image)
    # TODO: this is valid only for FRIDA
    kernel = np.array([[0.25, 0.50, 0.25]])

    # convolve RSS image
    convolved_data = convolve2d(image2d_rss_method1, kernel, boundary='fill', fillvalue=0, mode='same')

    # TODO: the second dimension in the following array should be 2*nslices
    # (check what to do for another IFU, like TARSIS)
    image3d_ifu_method1 = np.zeros((naxis1_detector.value, naxis2_ifu.value, naxis1_ifu.value))
    if verbose:
        print(f'(debug): {image3d_ifu_method1.shape=}')

    for islice in range(nslices):
        i1 = islice * 2
        j1 = islice * naxis1_ifu.value
        j2 = j1 + naxis1_ifu.value
        image3d_ifu_method1[:, i1, :] = convolved_data[j1:j2, :].T
        image3d_ifu_method1[:, i1+1, :] = convolved_data[j1:j2, :].T

    image3d_ifu_method1 /= 2
    if verbose:
        print(f'(debug): {np.sum(image2d_rss_method1)=}')
        print(f'(debug):      {np.sum(convolved_data)=}')
        print(f'(debug): {np.sum(image3d_ifu_method1)=}')

    return image3d_ifu_method1



#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import inspect

from numina.util.parser import parse_arg_line
from numina.datamodel import DataModel


class DataTypeBase(object):
    """Base class for input/output types of recipes.

    """

    def __init__(self, *args, **kwds):
        import numina.core.tagexpr as tagexpr

        datamodel = kwds.get('datamodel')

        if datamodel is not None:
            if inspect.isclass(datamodel):
                self.datamodel = datamodel()
            else:
                self.datamodel = datamodel
        else:
            self.datamodel = DataModel()

        my_tag_table = self.datamodel.query_attrs
        self.query_expr = tagexpr.ConstExprTrue

        if hasattr(self, '__tags__'):
            # FIXME:
            if isinstance(self.__tags__, list):
                objtags = [my_tag_table[t] for t in self.__tags__]
            elif isinstance(self.__tags__, dict):
                objtags = [t for t in self.__tags__.values()]
            else:
                raise TypeError('type not supported in tags')

            self.query_expr = tagexpr.query_expr_from_attr(objtags)
        if 'query_expr' in kwds:
            self.query_expr = kwds['query_expr']

        if 'tags' in kwds:
            # Create expresion from tags
            objtags = [my_tag_table[t] for t in kwds['tags']]
            self.query_expr = tagexpr.query_expr_from_attr(objtags)

        self.names_t = self.query_expr.tags()
        self.names_f = self.query_expr.fields()
        self.query_opts = []

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        pass

    def _datatype_dump(self, obj, where):
        return obj

    def _datatype_load(self, obj):
        return obj

    def query_constraints(self):
        from numina.core.query import Constraint
        return Constraint()

    @classmethod
    def isproduct(cls):
        """Check if the DataType is the product of a Recipe"""
        return False

    def __repr__(self):
        sclass = type(self).__name__
        return f"{sclass}()"

    def name(self):
        """Unique name of the datatype"""
        return self.__repr__()

    @classmethod
    def from_name(cls, name):
        # name is in the form Class(arg1=val1, arg2=val2)
        # find first (
        fp = name.find('(')
        sp = name.find(')')
        if (fp == -1) or (sp == -1) or (sp < fp):
            # paren not found
            klass = name
            kwargs = {}
        else:
            # parse things between parens
            klass = name[:fp]
            fargs = name[fp+1:sp]
            kwargs = parse_arg_line(fargs)

        if klass == cls.__name__:
            # create thing
            obj = cls.__new__(cls)
            obj.__init__(**kwargs)

            return obj
        else:
            raise TypeError(name)

    @staticmethod
    def create_db_info():
        """Create metadata structure"""
        result = {}
        result['instrument'] = ''
        result['uuid'] = ''
        result['tags'] = {}
        result['type'] = ''
        result['mode'] = ''
        result['observation_date'] = ""
        result['origin'] = {}
        return result

    def extract_db_info(self, obj, db_info_keys):
        """Extract metadata from serialized file"""
        result = self.create_db_info()
        return result

    def update_meta_info(self):
        """Extract my metadata"""
        result = self.create_db_info()
        return result

    def extract_tags(self, obj):
        """Create metadata structure"""
        return {}

    def tag_names(self):
        return self.names_t



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import warnings

import numina.util.convert as conv
from .keydef import FITSKeyExtractor


class DataModel:
    """Model of the Data being processed

    Parameters
    ==========
    name : str
        Name of the DataModel
    mappings: dict, optional
        A dictionary of str keys and values convertibles to FITSKeyExtractor

    """

    db_info_keys = [
        "instrument",
        "object",
        "observation_date",
        "uuid",
        "type",
        "mode",
        "exptime",
        "darktime",
        "insconf",
        "blckuuid",
        "quality_control",
    ]
    db_info_keys_extra = []
    query_attrs = {}
    meta_dinfo_headers = [
        "instrument",
        "object",
        "observation_date",
        "uuid",
        "type",
        "mode",
        "exptime",
        "darktime",
        "insconf",
        "blckuuid",
        "quality_control",
        "block_uuid",  # Alias
        "insconf_uuid",  # Alias
        "imgid",
    ]

    def __init__(self, name="UNKNOWN", mappings=None):

        self.name = name
        values = self.default_mappings()
        more = {} if mappings is None else mappings
        values.update(more)

        self.extractor_map = dict()
        self.extractor_map["fits"] = FITSKeyExtractor(values)
        self.extractor_map["json"] = None
        self.extractor = self.extractor_map["fits"]

    def default_mappings(self):
        return {
            "instrument": "INSTRUME",
            "object": "OBJECT",
            "observation_date": ("DATE-OBS", 0, conv.convert_date),
            "uuid": "uuid",
            "type": "numtype",
            "mode": "obsmode",
            "exptime": self.get_exptime,
            "darktime": self.get_darktime,
            "quality_control": ("NUMRQC", 0, conv.convert_qc),
            "insmode": ("INSMODE", "undefined"),
            "imgid": self.get_imgid,
            "insconf": "INSCONF",
            "blckuuid": lambda x: "1",
            "insconf_uuid": "insconf",  # Alias
            "block_uuid": "blckuuid",  # Alias
        }

    def get_data(self, img):
        """Obtain the primary data from the image"""
        return img["primary"].data

    def get_header(self, img):
        """Obtain the primary header from the image"""
        return img["primary"].header

    def get_variance(self, img):
        """Obtain the variance of the primary data from the image"""
        return img["variance"].data

    def get_imgid(self, img):
        """Obtain a unique identifier of the image.

        Parameters
        ----------
        img : astropy.io.fits.HDUList

        Returns
        -------
        str:
             Identification of the image

        """
        return get_imgid(img, prefix=False)

    def get_darktime(self, img):
        """Obtain DARKTIME"""
        header = self.get_header(img)
        if "DARKTIME" in header.keys():
            return header["DARKTIME"]
        else:
            return self.get_exptime(img)

    def get_exptime(self, img):
        """Obtain EXPTIME"""
        header = self.get_header(img)
        if "EXPTIME" in header.keys():
            etime = header["EXPTIME"]

        elif "EXPOSED" in header.keys():
            etime = header["EXPOSED"]
        else:
            etime = 1.0
        return etime

    def do_sky_correction(self, img):
        """Perform sky correction"""
        return True

    def gather_info(self, dframe):
        """Obtain a summary of information about the image."""
        with dframe.open() as hdulist:
            info = self.gather_info_hdu(hdulist)
        return info

    def gather_info_dframe(self, img):
        """Obtain a summary of information about the image."""
        return self.gather_info(img)

    def gather_info_hdu(self, hdulist):
        """Obtain a summary of information about the image."""
        values = {}
        values["n_ext"] = len(hdulist)
        extnames = [hdu.header.get("extname", "") for hdu in hdulist[1:]]
        values["name_ext"] = ["PRIMARY"] + extnames

        fits_extractor = self.extractor_map["fits"]
        for key in self.meta_dinfo_headers:
            values[key] = fits_extractor.extract(key, hdulist)

        return values

    def gather_info_oresult(self, ob):
        return [self.gather_info_dframe(f) for f in ob.images]

    def get_quality_control(self, img):
        """Obtain quality control flag from the image."""
        fits_extractor = self.extractor_map["fits"]
        return fits_extractor.extract("quality_control", img)


def get_imgid(img, prefix=True):
    """Obtain a unique identifier of the image.

    Parameters
    ----------
    img : astropy.io.fits.HDUList

    Returns
    -------
    str:
         Identification of the image

    """
    hdr = img[0].header
    try:
        return get_imgid_header(hdr, prefix=prefix)
    except ValueError:
        warnings.warn("no method to identify image", RuntimeWarning)
        value = repr(img)
        pre = "py:{}"
        base = "{}"
        if prefix:
            return pre.format(value)
        else:
            return base.format(value)


def get_imgid_header(hdr, prefix=True):
    """Obtain a unique identifier of the image.

    Parameters
    ----------
    hdr : astropy.io.fits.Header

    Returns
    -------
    str:
         Identification of the image

    """
    base = "{}"
    if "UUID" in hdr:
        pre = "uuid:{}"
        value = hdr["UUID"]
    elif "EMIRUUID" in hdr:  # EMIRISM
        pre = "uuid:{}"
        value = hdr["EMIRUUID"]
    elif "DATE-OBS" in hdr:
        pre = "dateobs:{}"
        value = hdr["DATE-OBS"]
    elif "TSUTC1" in hdr:  # EMIRISM
        pre = "tsutc:{:16.5f}"
        value = hdr["TSUTC1"]
    elif "checksum" in hdr:
        pre = "checksum:{}"
        value = hdr["checksum"]
    elif "filename" in hdr:
        pre = "file:{}"
        value = hdr["filename"]
    else:
        raise ValueError("no method to identity image")
    if prefix:
        return pre.format(value)
    else:
        return base.format(value)



#
# Copyright 2011-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DRP storage class"""

import warnings


class DrpBase(object):
    """Store DRPs, base class"""

    def __init__(self):
        pass

    def query_by_name(self, name):
        raise KeyError(name)

    def query_all(self):
        return {}

    @staticmethod
    def instrumentdrp_check(drpins, entryname):
        import numina.core.pipeline
        if isinstance(drpins, numina.core.pipeline.InstrumentDRP):
            if drpins.name == entryname:
                return True
            else:
                msg = f'Entry name "{entryname}" and DRP name "{drpins.name}" differ'
                warnings.warn(msg, RuntimeWarning)
                return False
        else:
            msg = f'Object {drpins!r} does not contain a valid DRP'
            warnings.warn(msg, RuntimeWarning)
            return False


class DrpGeneric(DrpBase):
    """Store DRPs.in a dictionary"""

    def __init__(self, drps=None):
        super(DrpGeneric, self).__init__()
        self.drps = drps

    def query_by_name(self, name):
        """Query DRPs in internal storage by name

        """
        return self._drps[name]

    def query_all(self):
        """Return all available DRPs in internal storage"""

        return self._drps

    @property
    def drps(self):
        return self._drps

    @drps.setter
    def drps(self, newdrps):
        self._drps = {} if newdrps is None else newdrps



from ..detector import DetectorBase


def test_detector_base():

    det_shape = (120, 240)
    dev = DetectorBase("detector", shape=det_shape)
    arr = dev.readout()
    assert arr.shape == det_shape



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import warnings
from astropy.io import fits

import numina.array


def get_hdu_shape(header):
    ndim = header['naxis']
    return tuple(header.get(f'NAXIS{i:d}') for i in range(1, ndim + 1))


def custom_slice_to_str(slc):
    if slc.step is None:
        return f'{slc.start:d}:{slc.stop:d}'
    else:
        return f'{slc.start:d}:{slc.stop:d}:{slc.step:d}'


def custom_region_to_str(region):
    jints = [custom_slice_to_str(slc) for slc in region]
    return '[' + ','.join(jints) + ']'


def resize_hdu(hdu, newshape, region, window=None, fill=0.0,
               scale=1, conserve=True, dtype=None):
    # from numina.array import resize_array

    basedata = hdu.data
    newdata = numina.array.resize_array(
        basedata, newshape, region, window=window,
        fill=fill, scale=scale, conserve=conserve,
        dtype=dtype
    )
    hdu.header['NVALREGI'] = (custom_region_to_str(region),
                              'Valid region of resized FITS')
    if window:
        hdu.header['OVALREGI'] = (custom_region_to_str(window),
                                  'Valid region of original FITS')
    newhdu = fits.PrimaryHDU(newdata, hdu.header)
    return newhdu


def resize_hdul(hdul, newshape, region, extensions=None, window=None,
                scale=1, fill=0.0, conserve=True):
    from numina.frame import resize_hdu
    if extensions is None:
        extensions = [0]

    nhdul = [None] * len(hdul)
    for ext, hdu in enumerate(hdul):
        if ext in extensions:
            nhdul[ext] = resize_hdu(hdu, newshape,
                                    region, fill=fill,
                                    window=window,
                                    scale=scale,
                                    conserve=conserve)
        else:
            nhdul[ext] = hdu
    return fits.HDUList(nhdul)


def resize_fits(fitsfile, newfilename, newshape, region, window=None,
                scale=1, fill=0.0, overwrite=True, conserve=True, dtype=None):

    close_on_exit = False
    if isinstance(fitsfile, str):
        hdulist = fits.open(fitsfile, mode='readonly')
        close_on_exit = True
    else:
        hdulist = fitsfile

    try:
        hdu = hdulist['primary']
        newhdu = resize_hdu(hdu, newshape, region, fill=fill,
                            window=window, scale=scale, conserve=conserve,
                            dtype=dtype)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            newhdu.writeto(newfilename, overwrite=overwrite)
    finally:
        if close_on_exit:
            hdulist.close()



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Numerical spline fit using different strategies."""

from lmfit import Minimizer, Parameters
import numpy as np
from scipy.interpolate import BSpline
from scipy.interpolate import LSQUnivariateSpline
from scipy.optimize import minimize


def fun_residuals(params, xnor, ynor, w, bbox, k, ext):
    """Compute fit residuals"""

    spl = LSQUnivariateSpline(
        x=xnor,
        y=ynor,
        t=[item.value for item in params.values()],
        w=w,
        bbox=bbox,
        k=k,
        ext=ext,
        check_finite=False
    )
    return spl.get_residual()


class AdaptiveLSQUnivariateSpline(LSQUnivariateSpline):
    """Extend scipy.interpolate.LSQUnivariateSpline.

    One-dimensional spline with explicit internal knots.

    This is actually a wrapper of
    `scipy.interpolate.LSQUnivariateSpline`
    with the addition of using adaptive knot location determined
    numerically (after normalising the x and y arrays before
    the minimisation process).

    Parameters
    ----------
    x : (N,) array_like
        Input dimension of data points -- must be increasing
    y : (N,) array_like
        Input dimension of data points
    t : (M,) array_like or int
        When integer it indicates the number of equidistant
        interior knots. When array_like it provides the location
        of the interior knots of the spline; must be in ascending
        order and::

         bbox[0] < t[0] < ... < t[-1] < bbox[-1]

    w : (N,) array_like, optional
        weights for spline fitting.  Must be positive.
        If None (default), weights are all equal.
    bbox : (2,) array_like, optional
        2-sequence specifying the boundary of the approximation
        interval. If None (default), ``bbox = [x[0], x[-1]]``.
    k : int, optional
        Degree of the smoothing spline.  Must be 1 <= `k` <= 5.
        Default is k=3, a cubic spline.
    ext : int or str, optional
        Controls the extrapolation mode for elements
        not in the interval defined by the knot sequence.

        * if ext=0 or 'extrapolate', return the extrapolated value.
        * if ext=1 or 'zeros', return 0
        * if ext=2 or 'raise', raise a ValueError
        * if ext=3 of 'const', return the boundary value.

        The default value is 0.

    check_finite : bool, optional
        Whether to check that the input arrays contain only finite
        numbers, that the x array is increasing and that the
        x and y arrays are 1-D and with the same length.
        Disabling may give a performance gain, but may
        result in problems (crashes, non-termination or non-sensical
        results) if the inputs do contain infinities or NaNs.
        Default is True.
    adaptive : bool, optional
        Whether to optimise knot location following the procedure
        described in Cardiel (2009); see:
        http://cdsads.u-strasbg.fr/abs/2009MNRAS.396..680C
        Default is True.
    tolerance : float, optional
        Tolerance for Nelder-Mead minimisation process.

    Attributes
    ----------
    _params : instance of Parameters()
        Initial parameters before minimisation.
    _result : Minimizer output
        Result of the minimisation process.

    See also
    --------
    LSQUnivariateSpline : Superclass

    """

    def __init__(self, x, y, t, w=None, bbox=(None, None),
                 k=3, ext=0, check_finite=False, adaptive=True,
                 tolerance=1E-7):
        """One-dimensional spline with explicit internal knots."""

        if check_finite:
            # check here the arrays instead of in the base class
            # (note that in the call to super(...).__init(...) the
            # parameter check_finite is set to False)
            w_finite = np.isfinite(x).all() if w is not None else True
            if not np.isfinite(x).all() or not np.isfinite(y).all() or \
                    not w_finite:
                raise ValueError('Input(s) must not contain '
                                 'NaNs or infs.')

            if np.asarray(x).ndim != 1:
                raise ValueError('x array must have dimension 1')

            if np.asarray(y).ndim != 1:
                raise ValueError('y array must have dimension 1')

            if np.asarray(x).shape != np.asarray(y).shape:
                raise ValueError('x and y arrays must have the same length')

            if not all(np.diff(x) > 0.0):
                raise ValueError('x array must be strictly increasing')

        # initial inner knot location (equidistant or fixed)
        try:
            nknots = int(t)
            if nknots > 0:
                xmin = x[0]
                xmax = x[-1]
                deltax = (xmax - xmin) / float(nknots + 1)
                xknot = np.zeros(nknots)
                for i in range(nknots):
                    xknot[i] = (xmin + float(i + 1) * deltax)
            else:
                xknot = np.array([])
        except (ValueError, TypeError):
            xknot = np.asarray(t)
            if check_finite:
                if not np.isfinite(xknot).all():
                    raise ValueError('Interior knots must not contain '
                                     'NaNs or infs.')
                if xknot.ndim != 1:
                    raise ValueError('t array must have dimension 1')
            nknots = len(xknot)

        # adaptive knots
        if nknots > 0 and adaptive:
            xknot_backup = xknot.copy()

            # normalise the x and y arrays to the [-1, +1] interval
            xmin = x[0]
            xmax = x[-1]
            ymin = np.min(y)
            ymax = np.max(y)
            bx = 2.0 / (xmax - xmin)
            cx = (xmin + xmax) / (xmax - xmin)
            by = 2.0 / (ymax - ymin)
            cy = (ymin + ymax) / (ymax - ymin)
            xnor = bx * np.asarray(x) - cx
            ynor = by * np.asarray(y) - cy
            xknotnor = bx * xknot - cx
            params = Parameters()
            for i in range(nknots):
                if i == 0:
                    xminknot = xnor[0]
                    if nknots == 1:
                        xmaxknot = xnor[-1]
                    else:
                        xmaxknot = (xknotnor[i] + xknotnor[i+1]) / 2.0
                elif i == nknots - 1:
                    xminknot = (xknotnor[i-1] + xknotnor[i]) / 2.0
                    xmaxknot = xnor[-1]
                else:
                    xminknot = (xknotnor[i-1] + xknotnor[i]) / 2.0
                    xmaxknot = (xknotnor[i] + xknotnor[i+1]) / 2.0
                params.add(
                    name=f'xknot{i:03d}',
                    value=xknotnor[i],
                    min=xminknot,
                    max=xmaxknot,
                    vary=True
                )
            self._params = params.copy()
            fitter = Minimizer(
                userfcn=fun_residuals,
                params=params,
                fcn_args=(xnor, ynor, w, bbox, k, ext)
            )
            try:
                self._result = fitter.scalar_minimize(
                    method='Nelder-Mead',
                    tol=tolerance
                )
                xknot = [item.value for item in self._result.params.values()]
                xknot = (np.asarray(xknot) + cx) / bx
            except ValueError:
                print('Error when fitting adaptive splines. '
                      'Reverting to initial knot location.')
                xknot = xknot_backup.copy()
                self._result = None
        else:
            self._params = None
            self._result = None

        # final fit
        super(AdaptiveLSQUnivariateSpline, self).__init__(
            x=x,
            y=y,
            t=xknot,
            w=w,
            bbox=bbox,
            k=k,
            ext=ext,
            check_finite=False
        )

    def get_params(self):
        """Return initial parameters for minimisation process."""

        return self._params

    def get_result(self):
        """Return result of minimisation process."""

        return self._result


def spline_positive_derivative(x, y, w=None, n_total_knots=2, derivative_threshold=1e-6):
    """Fit a spline to data with positive derivative constraint.

    Note that this function uses a B-spline basis and optimizes
    the coefficients to minimize the weighted mean squared error
    while ensuring that the derivative of the spline is strictly
    positive at all points.

    The knots are evenly spaced across the range of `x`, and the
    degree of the spline is set to 3 (cubic spline).

    Parameters
    ----------
    x : array_like
        Input x-coordinates of the data points.
    y : array_like
        Input y-coordinates of the data points.
    w : array_like, optional
        Weights for the data points. If None, all points are equally
        weighted.
    n_total_knots : int, optional
        Total number of knots to use in the spline fit. Default is 2.
    derivative_threshold : float, optional
        Minimum value for the derivative of the spline to ensure it is
        strictly positive.

    Returns
    -------
    spline_fit : BSpline
        The fitted B-spline object.
    knots : array_like
        The knot vector used in the spline fit.
    """
    degree = 3

    if n_total_knots < 2:
        raise ValueError("n_total_knots must be at least 2.")

    if w is None:
        w = np.ones_like(x, dtype=float)

    knots = np.linspace(x.min(), x.max(), n_total_knots)

    # Full knot vector
    t = np.concatenate(([x.min()] * degree, knots, [x.max()] * degree))

    # Number of basis functions
    n_bases = len(t) - degree - 1

    # Evaluate B-spline basis functions at x
    def bspline_design_matrix(x, t, k):
        n = len(t) - k - 1
        B = np.zeros((len(x), n))
        for i in range(n):
            coeff = np.zeros(n)
            coeff[i] = 1
            B[:, i] = BSpline(t, coeff, k)(x)
        return B

    B = bspline_design_matrix(x, t, degree)

    # Define weighted loss function (weighted mean squared error)
    def weighted_loss(c):
        y_pred = B @ c
        return np.mean(w * (y - y_pred)**2)

    # Define constraint: derivative must be strictly positive
    def make_constraints():
        x_dense = np.linspace(x.min(), x.max(), 100)
        return [
            {
                'type': 'ineq',
                'fun': lambda c, xi=xi: BSpline(t, c, degree).derivative()(xi) - derivative_threshold
            }
            for xi in x_dense
        ]

    # Initial guess for coefficients
    c0 = np.ones(n_bases)

    # Perform constrained optimization
    constraints = make_constraints()
    result = minimize(weighted_loss, c0, constraints=constraints, method='SLSQP')

    # Evaluate fitted spline
    c_opt = result.x
    spline_fit = BSpline(t, c_opt, degree)

    # Return the fitted spline and knots
    return spline_fit, knots



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import math
from astropy.modeling.functional_models import Gaussian2D, Moffat2D


class AtmosphereModel(object):

    def __init__(self, twilight, nightsky, seeing, extinction, refraction):
        self.tw_interp = twilight
        self.ng_interp = nightsky
        self.ext_interp = extinction
        self.seeing = seeing
        self.refraction_model = refraction

    def twilight_spectrum(self, wl_in):
        """Twilight spectrum"""
        return self.tw_interp(wl_in)

    def night_spectrum(self, wl_in):
        """Night spectrum"""
        return self.ng_interp(wl_in)

    def extinction(self, wl_in):
        """Night extinction"""
        return self.ext_interp(wl_in)

    def refraction(self, z, wl, ref):
        return self.refraction_model.refraction(z, wl, ref)


class SeeingSizeModel(object):
    def __init__(self, wl, r0):
        self._r0 = r0
        self._wl0 = wl

    def fwhm(self, wl, zd):
        return 1.2 * wl / self.r0(wl, zd)

    def r0(self, wl, zd):
        return self._r0 * (wl / self._wl0) ** 1.2 * math.cos(zd)

    def profile(self, fwhm):
        return generate_gaussian_profile(fwhm)


class ConstSeeing(object):
    def __init__(self, seeing):
        self._s = seeing

    def fwhm(self, wl, zd):
        return self._s

    def profile(self, fwhm):
        return generate_gaussian_profile(fwhm)


def generate_gaussian_profile(seeing_fwhm):
    """Generate a normalized Gaussian profile from its FWHM"""
    FWHM_G = 2 * math.sqrt(2 * math.log(2))
    sigma = seeing_fwhm / FWHM_G
    amplitude = 1.0 / (2 * math.pi * sigma * sigma)
    seeing_model = Gaussian2D(amplitude=amplitude,
                              x_mean=0.0,
                              y_mean=0.0,
                              x_stddev=sigma,
                              y_stddev=sigma)
    return seeing_model


def generate_moffat_profile(seeing_fwhm, alpha):
    """Generate a normalized Moffat profile from its FWHM and alpha"""

    scale = 2 * math.sqrt(2**(1.0 / alpha) - 1)
    gamma = seeing_fwhm / scale
    amplitude = 1.0 / math.pi * (alpha - 1) / gamma**2
    seeing_model = Moffat2D(amplitude=amplitude,
                            x_mean=0.0,
                            y_mean=0.0,
                            gamma=gamma,
                            alpha=alpha)
    return seeing_model


def generate_lorentz_profile(seeing_fwhm):
    """Generate a normalized Lorent profile from its FWHM"""

    return generate_moffat_profile(seeing_fwhm, alpha=1.5)



#
# Copyright 2011-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Build a LoadableDRP from a yaml file"""

import pkgutil
import importlib
from io import StringIO

import yaml

from numina.util.objimport import import_object
from .pipeline import ObservingMode
from .pipeline import Pipeline
from .pipeline import InstrumentDRP
# from .instrument.insconf import InstrumentConfiguration, instrument_loader
from .pipeline import ProductEntry
from .query import ResultOf
from .taggers import get_tags_from_full_ob


def check_section(node, section, keys=None):
    """Validate keys in a section"""
    if keys:
        for key in keys:
            if key not in node:
                raise ValueError(
                    f'Missing key {key!r} inside {section!r} node')


def drp_load(package, resource, confclass=None):
    """Load the DRPS from a resource file."""
    data = pkgutil.get_data(package, resource)
    return drp_load_data(package, data, confclass=confclass)


def drp_load_data(package, data, confclass=None):
    """Load the DRPS from data."""
    drpdict = yaml.safe_load(data)
    ins = load_instrument(package, drpdict, confclass=confclass)
    if ins.version == 'undefined':
        pkg = importlib.import_module(package)
        ins.version = getattr(pkg, '__version__', 'undefined')
    return ins


def load_modes(node, confclass=None):
    """Load all observing modes"""
    if isinstance(node, list):
        values = [load_mode(child, confclass) for child in node]
        keys = [mode.key for mode in values]
        return dict(zip(keys, values))
    elif isinstance(node, dict):
        values = {key: load_mode(child) for key, child in node}
        return values
    else:
        raise NotImplementedError


def load_mode(node, confclass=None):
    """Load one observing mdode"""
    obs_mode = ObservingMode()
    if confclass is not None:
        node = confclass.mode_loader(node)
    obs_mode.__dict__.update(node)

    # handle validator
    load_mode_validator(obs_mode, node)

    # handle builder
    load_mode_builder(obs_mode, node)

    # handle tagger:
    load_mode_tagger(obs_mode, node)

    return obs_mode


def load_mode_tagger(obs_mode, node):
    """Load observing mode OB tagger"""

    # handle tagger:
    ntagger = node.get('tagger')

    if ntagger is None:
        pass
    elif isinstance(ntagger, list):

        def full_tagger(obsres):
            return get_tags_from_full_ob(obsres, reqtags=ntagger)

        obs_mode.tagger = full_tagger
    elif isinstance(ntagger, str):
        # load function
        obs_mode.tagger = import_object(ntagger)
    else:
        raise TypeError('tagger must be None, a list or a string')

    return obs_mode


def load_mode_builder(obs_mode, node):
    """Load observing mode OB builder"""

    # Check 'builder' and 'builder_options'
    nval1 = node.get('builder')

    if nval1 is not None:
        if isinstance(nval1, str):
            # override method
            newmethod = import_object(nval1)
            obs_mode.build_ob = newmethod.__get__(obs_mode)
        else:
            raise TypeError('builder must be None or a string')
    else:
        nval2 = node.get('builder_options')

        if nval2 is not None:
            if isinstance(nval2, list):
                for opt_dict in nval2:

                    if 'result_of' in opt_dict:
                        fields = opt_dict['result_of']
                        obs_mode.build_ob_options = ResultOf(**fields)
                        break
            else:
                raise TypeError('builder_options must be None or a list')

    return obs_mode


def load_mode_validator(obs_mode, node):
    """Load observing mode validator"""

    nval = node.get('validator')

    if nval is None:
        pass
    elif isinstance(nval, str):
        # load function
        obs_mode.validator = import_object(nval)
    else:
        raise TypeError('validator must be None or a string')

    return obs_mode


def load_pipelines(instrument, node):
    keys = ['default']
    check_section(node, 'pipelines', keys=keys)

    pipelines = {}
    for key in node:
        pipelines[key] = load_pipeline(instrument, key, node[key])
    return pipelines


def load_confs(package, node, confclass=None):
    import numina.instrument.assembly as asbl
    keys = []
    check_section(node, 'configurations', keys=keys)

    path = node.get('path')
    if path:
        modpath = path
    else:
        modpath = f"{package}.instrument.configs"

    if confclass is None:
        _loader = DefaultLoader(modpath=modpath)  # noqa: F841

    tagger = node.get('tagger')
    if tagger:
        ins_tagger = import_object(tagger)
    else:
        ins_tagger = None

    confs = asbl.load_paths_store([modpath])
    return confs, ins_tagger, modpath


def load_pipeline(instrument, name, node):

    keys = ['recipes', 'version']
    check_section(node, 'pipeline', keys=keys)

    recipes = load_base("recipes", node['recipes'])
    if 'products' in node:
        products = load_base("products", node['products'])
    else:
        products = {}
    if 'provides' in node:
        provides = load_prods(node['provides'], recipes.keys())
    else:
        provides = []
    version = node['version']
    return Pipeline(instrument, name, recipes, version=version,
                    products=products, provides=provides)


def load_recipe(name, node):

    recipe = {'class': ''}

    keys = ['class']
    if isinstance(node, dict):
        check_section(node, name, keys=keys)
        recipe = node
    else:
        recipe['class'] = node
    if 'args' in recipe:
        recipe['args'] = tuple(recipe['args'])
    if 'links' in recipe:
        recipe['links'] = load_link(recipe['links'])
    return recipe


def load_link(node):
    """Build a ResultOf query from dict"""
    result = {}
    for key, opt_dict in node.items():

        if 'result_of' in opt_dict:
            fields = opt_dict['result_of']
            result[key] = ResultOf(**fields)
        else:
            pass
    return result


def load_base(name, node):

    # keys = ['recipes', 'version']
    # check_section(node, 'pipeline', keys=keys)
    recipes = {}
    for key in node:
        recipes[key] = load_recipe(key, node[key])
    # print(recipes)
    return recipes


def load_prods(node, allmodes):
    result = []
    for entry in node:
        name = entry['name']
        mode_name = entry['mode']
        field = entry['field']
        for mode_key in allmodes:
            if mode_key == mode_name:
                prod = ProductEntry(name, mode_key, field)
                result.append(prod)
                break
        else:
            # Undefined mode
            pass

    return result


def load_instrument(package, node, confclass=None):
    # Verify keys...
    keys = ['name', 'configurations', 'modes', 'pipelines']
    check_section(node, 'root', keys=keys)

    # name = node['name']
    pipe_node = node['pipelines']
    mode_node = node['modes']
    conf_node = node['configurations']

    trans = {'name': node['name'], 'version': 'undefined'}
    if 'datamodel' in node:
        trans['datamodel'] = import_object(node['datamodel'])
    else:
        trans['datamodel'] = None
    if 'version' in node:
        trans['version'] = node['version']
    trans['pipelines'] = load_pipelines(node['name'], pipe_node)
    trans['modes'] = load_modes(mode_node, confclass)
    confs, custom_selector, modpath = load_confs(
        package, conf_node, confclass=confclass)
    # trans['configurations'] = confs
    trans['configurations'] = confs
    ins = InstrumentDRP(**trans)
    # idiom to add a bound method
    if custom_selector:
        ins.select_configuration = custom_selector.__get__(ins)
    # Add package name
    ins.package = package
    # Add profile path
    ins.profiles = modpath
    return ins


class DefaultLoader(object):
    def __init__(self, modpath):
        self.modpath = modpath

    def build_component_fp(self, key):
        fname = f'component-{key}.json'
        return self.build_type_fp(fname)

    def build_instrument_fp(self, key):
        fname = f'instrument-{key}.json'
        return self.build_type_fp(fname)

    def build_type_fp(self, fname):
        data = pkgutil.get_data(self.modpath, fname)
        fcomp = StringIO(data.decode('utf-8'))
        return fcomp



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Apply integer (pixel) offsets to an image"""

import argparse
import astropy.io.fits as fits
import numpy as np
import sys


def apply_integer_offsets(image2d, offx, offy):
    """Apply global (integer) offsets to image.

    Parameters
    ----------
    image2d : numpy array
        Input image
    offx : int
        Offset in the X direction (must be integer).
    offy : int
        Offset in the Y direction (must be integer).

    Returns
    -------
    image2d_shifted : numpy array
        Shifted image

    """

    # protections
    if not isinstance(offx, int) or not isinstance(offy, int):
        raise ValueError(f'Invalid non-integer offsets: {offx=} of type {type(offx)}, {offy=} of type {type(offy)}')

    # image dimensions
    naxis2, naxis1 = image2d.shape

    # initialize output image
    image2d_shifted = np.zeros((naxis2, naxis1))

    # handle negative and positive shifts accordingly
    def non(s): return s if s < 0 else None  # noqa: E731
    def mom(s): return max(0, s)  # noqa: E731

    # shift image
    image2d_shifted[mom(offy):non(offy), mom(offx):non(offx)] = \
        image2d[mom(-offy):non(-offy), mom(-offx):non(-offx)]

    # return shifted image
    return image2d_shifted


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()
    # required parameters
    parser.add_argument("infile",
                        help="Input FITS image",
                        type=argparse.FileType('rb'))
    parser.add_argument("outfile",
                        help="Output FITS image",
                        type=argparse.FileType('wb'))
    # optional arguments
    parser.add_argument("--offx",
                        help="Offset in the X direction (integer)",
                        default=0, type=int)
    parser.add_argument("--offy",
                        help="Offset in the Y direction (integer)",
                        default=0, type=int)
    parser.add_argument("--extension",
                        help="Extension number in FITS image (0=first "
                             "extension)",
                        default=0, type=int)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    # read input FITS file
    with fits.open(args.infile) as hdulist:
        image2d = hdulist[args.extension].data
        hdulist[args.extension].data = apply_integer_offsets(
            image2d=image2d,
            offx=args.offx,
            offy=args.offy
        )

    # save output FITS file
    hdulist.writeto(args.outfile, overwrite=True)


if __name__ == "__main__":

    main()




import pytest

from ..dataholders import Parameter

from numina.core.validator import as_list, only_positive
from numina.types.datatype import PlainPythonType
from numina.core.recipeinout import RecipeInput
from numina.exceptions import ValidationError
from numina.types.datatype import ListOfType


def test_requirement_in():

    value1 = 1
    value2 = 100

    class RR(RecipeInput):
        some = Parameter(value1, '')

    rr = RR()

    assert rr.some == value1

    class RR(RecipeInput):
        some = Parameter(value1, '')

    rr = RR(some=value2)

    assert rr.some == value2


def test_param_internal():

    value1 = 1

    class RR(RecipeInput):
        some = Parameter(value1, '')

    assert isinstance(RR.some, Parameter)
    assert isinstance(RR.some.type, PlainPythonType)


def validator_fun1(value):
    if value > 50:
        raise ValidationError
    return value


def test_param_check():

    value1 = 1
    value2 = 100
    value3 = 20

    class RR(RecipeInput):
        some = Parameter(value1, '', validator=validator_fun1)

    with pytest.raises(ValidationError):
        RR(some=value2)

    rr = RR(some=value3)

    assert rr.some == value3


def test_param_choices1():

    value1 = 1
    choices = [1, 2, 3, 4]
    value2 = 2

    class RR(RecipeInput):
        some = Parameter(value1, '', choices=choices)

    rr = RR(some=value2)

    assert rr.some == value2


def test_param_choices2():

    value1 = 1
    choices = [1, 2, 3, 4]
    value2 = 8

    class RR(RecipeInput):
        some = Parameter(value1, '', choices=choices)

    with pytest.raises(ValidationError):
        RR(some=value2)


def test_param_dest():

    default = 34.3

    class RR(RecipeInput):
        some = Parameter(default, '', destination='other')

    assert hasattr(RR, 'other')
    assert not hasattr(RR, 'some')

    rr = RR()
    assert hasattr(rr, 'other')
    assert not hasattr(rr, 'some')

    rr = RR(other=default)
    assert hasattr(rr, 'other')
    assert not hasattr(rr, 'some')


def test_param_accept_scalar():
    """Test accept_scalar argument"""
    some = Parameter([1], 'Convert scalar to list', accept_scalar=True)
    result = some.convert(1)
    assert result == [1]

    assert isinstance(some.type, ListOfType)
    assert isinstance(some.type.node_type, PlainPythonType)
    assert some.type.internal_type is list


def test_param_as_list1():
    """Test as_list argument"""
    some = Parameter(1, 'Convert scalar to list', as_list=True)
    result = some.convert([1, 2, 3])
    assert result == [1, 2, 3]

    result = some.convert(1)
    assert result == [1]


def test_param_as_list2():
    """Test list of tuples"""
    some = Parameter([(0, 0)], 'List of coordinates')
    result = some.convert([(3, 4), (1, 1)])
    assert result == [[3, 4], [1, 1]]


@pytest.mark.xfail
def test_param_as_list3():
    """Test list of tuples"""
    some = Parameter([(0, 0)], 'List of coordinates')
    result = some.convert([(3, 4), (1, 1)])
    assert result == [(3, 4), (1, 1)]


@pytest.mark.parametrize("nelem, allowed, not_allowed", [
    ('*', [[], [1], [2, 3, 4]], []),
    ('+', [[1], [2, 3, 4]], [[]]),
    (1, [[1]], [[], [3.1, 34.0, 4]])
])
def test_param_as_list4(nelem, allowed, not_allowed):
    """Test list of tuples"""
    some = Parameter(5, 'List of coordinates', nelem=nelem)
    for obj in allowed:
        assert obj == some.convert(obj)

    for obj in not_allowed:
        with pytest.raises(ValidationError):
            some.convert(obj)


def test_param_custom_validator1():
    """Test accept_scalar argument"""
    some = Parameter(1, 'Validate', validator=only_positive)

    with pytest.raises(ValidationError):
        some.convert(-11)


def test_param_custom_validator2():
    """Test accept_scalar argument"""
    some = Parameter(1, 'Validate', as_list=True, validator=only_positive)

    with pytest.raises(ValidationError):
        some.convert(-11)

    with pytest.raises(ValidationError):
        some.convert([-11])


def test_param_custom_validator3():
    """Test accept_scalar argument"""
    some = Parameter([1], 'Validate', validator=as_list(only_positive))

    with pytest.raises(ValidationError):
        some.convert([-33])


def test_param_custom_validator4():
    """Test accept_scalar argument"""
    some = Parameter([1], 'Validate', accept_scalar=True,
                     validator=as_list(only_positive))

    with pytest.raises(ValidationError):
        some.convert(-11)

    with pytest.raises(ValidationError):
        some.convert([50, -22])


def test_param_default1():
    """Test accept_scalar argument"""
    some = Parameter([1], 'Validate')

    assert some.default_value() == [1]


def test_param_default2():
    """Test accept_scalar argument"""
    some = Parameter(1, 'Validate', as_list=True)

    assert some.default_value() == [1]


def test_empty_list_escalar():
    some = Parameter([], 'Some', accept_scalar=True)
    value = some.convert(1)
    assert value == [1]


def test_empty_list_iter():
    some = Parameter([], 'Some')
    value = some.convert([2])
    assert value == [2]

    value = some.convert(["A", 34.2])
    assert value == ["A", 34.2]


def test_param_bool1():
    """Test bool argument"""

    value1 = True
    value2 = False

    class RR(RecipeInput):
        some = Parameter(value1, '')

    rr = RR()

    assert rr.some == value1

    class RR(RecipeInput):
        some = Parameter(value2, '')

    rr = RR(some=value2)

    assert rr.some == value2



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Store HDU from different files into a single FITS file with extensions.
"""
import argparse
from astropy.io import fits
import sys

from .ctext import ctext


def stack_hdus(input_list, output_filename, verbose=False):
    """Store HDU from different files into a single FITS file with extensions.

    Parameters
    ----------
    input_list : str
        Input text file with the name of the FITS files
        whose HDU will be stacked together in a single FITS file.
    output_filename : str
        Output FITS file with the stacked HDUs.
    verbose : bool
        If True, display additional information.
    """

    # read input file with list of input FITS images
    with open(input_list) as f:
        file_content = f.read().splitlines()

    hdul_stack = fits.HDUList()
    nstack = 0
    for line in file_content:
        if len(line) > 0:
            if line[0] not in ['#']:
                file_and_extension = line.split(',')
                fname = file_and_extension[0]
                if len(file_and_extension) == 1:
                    extname = 'PRIMARY'
                elif len(file_and_extension) == 2:
                    extname = file_and_extension[1].upper()
                else:
                    raise ValueError(f"Unexpected file, extension: {file_and_extension}")
                if verbose:
                    print(f'* Reading: {fname} -> HDU: {extname}')
                with fits.open(fname) as hdul:
                    if extname not in hdul:
                        raise ValueError(f'Expected {extname} extension not found')
                    data = hdul[extname].data
                    if nstack == 0:
                        hdu = fits.PrimaryHDU(data)
                    else:
                        hdu = fits.ImageHDU(data)
                    hdul_stack.append(hdu)

    if verbose:
        print(f"Saving {output_filename}")
    hdul_stack.writeto(output_filename, overwrite=True)


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input_list",
                        help="TXT file with list of images whose HDU will be stacked", type=str)
    parser.add_argument('output_filename',
                        help='filename of output FITS image', type=str)

    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_list = args.input_list
    output_filename = args.output_filename
    verbose = args.verbose

    stack_hdus(input_list, output_filename, verbose)


if __name__ == '__main__':
    main()


#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Constants for the numina package."""


# Factor to obtain FWHM in Gaussian profile from sigma
FWHM_G = 2.3548200450309493


# Square root of two pi
M_SQRT_2_PI = 2.5066282746310005024157652848110452530069867406099383166299



import numpy as np
from scipy.ndimage import gaussian_filter1d

from ..broadsp_gaussian_linearwv import apply_gaussian_broadening_linearwv


def test_apply_constant_sigma_in_pixels():

    rng = np.random.default_rng(1234)

    # generate fake spectrum
    naxis1 = 1000
    flux = np.zeros(naxis1)
    border = 10
    nlines = 15
    ipix = rng.uniform(low=border, high=naxis1-border, size=nlines).astype(int)
    for i in (ipix):
        flux[i] = rng.uniform(low=0.2, high=1.0, size=1)[0]
    flux = gaussian_filter1d(flux, 3)
    flux = 1 - flux

    # compute the effect of a gaussian broadining with a kernel of fixed
    # size (in pixels)
    method1 = gaussian_filter1d(
        input=flux,
        sigma=5.0,
        truncate=4.0,
        mode='constant'
    )
    method2 = apply_gaussian_broadening_linearwv(
        flux=flux,
        sigma_pix=np.ones_like(flux) * 5.0,
        tsigma=4.0
    )

    # compare results from both methods (should be identical)
    assert np.allclose(method1, method2)



#
# Copyright 2016-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Combination routines"""

import datetime
import logging
import uuid
import contextlib

from astropy.io import fits

from numina.array import combine
from numina.datamodel import get_imgid


def basic_processing_with_combination(
        rinput, reduction_flows,
        method=combine.mean, method_kwargs=None,
        errors=True, prolog=None):

    return basic_processing_with_combination_frames(
        rinput.obresult.frames, reduction_flows,
        method=method, method_kwargs=method_kwargs,
        errors=errors, prolog=prolog
    )


def basic_processing_with_combination_frames(
        frames, reduction_flows,
        method=combine.mean, method_kwargs=None,
        errors=True, prolog=None):

    result = combine_frames(
        frames, method=method, method_kwargs=method_kwargs,
        errors=errors, prolog=prolog
    )

    if isinstance(reduction_flows, list):
        # FIXME: handling a list, we use only the first element
        reduction_flow = reduction_flows[0]
    else:
        reduction_flow = reduction_flows

    hdulist = reduction_flow(result)

    return hdulist


def combine_frames(frames, method=combine.mean, method_kwargs=None, errors=True, prolog=None):
    """

    Parameters
    ----------
    frames
    method
    method_kwargs
    errors
    prolog

    Returns
    -------

    """

    with contextlib.ExitStack() as stack:
        hduls = [stack.enter_context(dframe.open()) for dframe in frames]
        result = combine_imgs(
            hduls, method=method, method_kwargs=method_kwargs,
            errors=errors, prolog=prolog
        )

    return result


def combine_imgs(hduls, method=combine.mean, method_kwargs=None,
                 errors=True, prolog=None, crmasks=None, use_lamedian=False):
    """

    Parameters
    ----------
    hduls
    method
    method_kwargs
    errors
    prolog
    crmasks

    Returns
    -------

    """

    _logger = logging.getLogger(__name__)

    cnum = len(hduls)
    if cnum == 0:
        raise ValueError('number of HDUList == 0')

    first_image = hduls[0]
    base_header = first_image[0].header.copy()
    last_header = hduls[-1][0].header.copy()

    method_kwargs = method_kwargs or {}
    if 'dtype' not in method_kwargs:
        method_kwargs['dtype'] = 'float32'

    _logger.info(f"stacking {cnum:d} images using '{method.__name__}'")
    if method.__name__ in ['mediancr', 'meancrt', 'meancr']:
        combined_data = method([d[0].data for d in hduls], crmasks=crmasks, **method_kwargs)
    else:
        combined_data = method([d[0].data for d in hduls], **method_kwargs)

    hdu = fits.PrimaryHDU(combined_data[0], header=base_header)
    _logger.debug('update result header')
    if prolog:
        _logger.debug('write prolog')
        hdu.header['history'] = prolog
    hdu.header['history'] = f"Combined {cnum:d} images using '{method.__name__}'"
    t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
    hdu.header['history'] = f'Combination time {t_str}'

    for img in hduls:
        hdu.header['history'] = f"Image {get_imgid(img)}"

    prevnum = base_header.get('NUM-NCOM', 1)
    hdu.header['NUM-NCOM'] = prevnum * cnum
    hdu.header['UUID'] = str(uuid.uuid1())

    # Copy extensions and then append 'variance' and 'map'
    result = fits.HDUList([hdu])
    for hdu in first_image[1:]:
        result.append(hdu.copy())

    # Headers of last image, this is an EMIRISM
    if 'TSUTC2' in hdu.header:
        hdu.header['TSUTC2'] = last_header['TSUTC2']
    # Append error extensions
    if errors:
        varhdu = fits.ImageHDU(combined_data[1], name='VARIANCE')
        result.append(varhdu)
        num = fits.ImageHDU(combined_data[2].astype('int16'), name='MAP')
        result.append(num)

    return result


def main(args=None):
    import argparse

    parser = argparse.ArgumentParser(prog='combine')
    parser.add_argument('-o', '--output', default='combined.fits')
    parser.add_argument('-e', '--errors', default=False, action='store_true')
    parser.add_argument('--method', default='mean', choices=['mean', 'median'])
    parser.add_argument('image', nargs='+')
    args = parser.parse_args(args)

    if args.method == 'mean':
        method = combine.mean
    elif args.method == 'median':
        method = combine.median
    else:
        raise ValueError(f'wrong method {args.method}')

    errors = args.errors
    with contextlib.ExitStack() as stack:
        hduls = [stack.enter_context(fits.open(fname)) for fname in args.image]
        result = combine_imgs(hduls, method=method, errors=errors, prolog=None)

    result.writeto(args.output, overwrite=True)


if __name__ == '__main__':
    main()



#
# Copyright 2008-2016 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

if __name__ == '__main__':

    from numina.user.cli import main

    main()




import importlib.metadata
import pkgutil

import backports.entry_points_selectable

from numina.drps.drpsystem import DrpSystem
from numina.core.pipeline import InstrumentDRP, Pipeline


def assert_valid_instrument(instrument):
    assert isinstance(instrument, InstrumentDRP)

    pipes = instrument.pipelines
    assert 'default' in pipes
    for k, v in pipes.items():
        assert k == v.name
        assert isinstance(v, Pipeline)


def test_fake_pipeline(monkeypatch):

    def mockreturn(group=None):

        def fake_loader():
            confs = dict()
            modes = dict()
            pipelines = {'default': Pipeline('default', {}, 1)}
            fake = InstrumentDRP('FAKE', confs, modes, pipelines)
            return fake

        ep = importlib.metadata.EntryPoint(
            name='fake', value='fake.loader', group='numina.pipelines.1')
        monkeypatch.setattr(ep, 'load', lambda: fake_loader)
        return [ep]

    monkeypatch.setattr(backports.entry_points_selectable,
                        'entry_points', mockreturn)

    alldrps = DrpSystem().load().query_all()
    for k, v in alldrps.items():
        assert_valid_instrument(v)


def test_fake_pipeline_alt(drpmocker):

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')

    drpmocker.add_drp('TEST1', drpdata1)

    mydrp = DrpSystem().load().query_by_name('TEST1')
    assert mydrp is not None

    assert_valid_instrument(mydrp)


def test_fake_pipeline_alt2(drpmocker):

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')

    ob_to_test = """
    id: 4
    mode: bias
    instrument: TEST1
    images:
     - ThAr_LR-U.fits
    """

    drpmocker.add_drp('TEST1', drpdata1)

    import yaml
    from numina.core.oresult import obsres_from_dict
    from numina.tests.recipes import BiasRecipe

    oblock = obsres_from_dict(yaml.safe_load(ob_to_test))

    drp = DrpSystem().load().query_by_name(oblock.instrument)

    assert drp is not None

    assert_valid_instrument(drp)

    this_pipeline = drp.pipelines[oblock.pipeline]
    expected = 'numina.tests.recipes.BiasRecipe'
    assert this_pipeline.get_recipe(oblock.mode) == expected

    recipe = this_pipeline.get_recipe_object(oblock.mode)
    assert isinstance(recipe, BiasRecipe)

    assert recipe.instrument == "TEST1"
    assert recipe.mode == "bias"
    assert recipe.simulate_error is True



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
import numpy as np


def save_image2d_detector_method0(
        header_keys,
        image2d_detector_method0,
        prefix_intermediate_fits,
        bitpix
):
    """Save the two 2D images: RSS and detector.

    Parameters
    ----------
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the default keywords.
    image2d_detector_method0 : `~numpy.ndarray`
        2D array containing the detector image.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    bitpix : int, optional
        BITPIX value for the FITS file.
    """

    if len(prefix_intermediate_fits) > 0:
        # --------------------------------------
        # spectroscopic 2D image in the detector
        # --------------------------------------
        if bitpix == 16:
            # avoid overflow
            image2d_detector_method0[image2d_detector_method0 > 65535] = 65535
            # round to integer and save as BITPIX=16 (unsigned short)
            hdu = fits.PrimaryHDU(np.round(image2d_detector_method0).astype(np.uint16))
        else:
            raise ValueError(f'Unsupported BITPIX value: {bitpix}')
        pos0 = len(hdu.header) - 1
        hdu.header.update(header_keys)
        hdu.header.insert(
            pos0, ('COMMENT', "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ('COMMENT', "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"))
        hdul = fits.HDUList([hdu])
        outfile = f'{prefix_intermediate_fits}_detector_2D_method0.fits'
        print(f'Saving file: {outfile}')
        hdul.writeto(outfile, overwrite='yes')



#
# Copyright 2011-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DRP related classes"""

import warnings
import logging

import numina.core.query
import numina.core.deptree
import numina.util.objimport
import numina.util.parser
import numina.datamodel


class Pipeline:
    """Base class for pipelines."""

    def __init__(self, instrument, name, recipes, version=1, products=None, provides=None):
        self.instrument = instrument
        self.name = name
        self.recipes = recipes
        self.products = {} if products is None else products
        self.version = version

        # Query by different keys
        self._provides_by_p = {}
        self._provides_by_r = {}

        provides_ = [] if provides is None else provides
        for k in provides_:
            self._provides_by_p[k.name] = k
            self._provides_by_r[k.mode] = k

    def get_recipe(self, mode):
        node = self.recipes[mode]
        return node['class']

    def _get_base_class(self, entry):

        recipe_fqn = entry['class']
        return numina.util.objimport.import_object(recipe_fqn)

    def _get_base_object(self, entry):

        Cls = self._get_base_class(entry)

        args = entry.get('args', ())
        kwargs = entry.get('kwargs', {})
        links = entry.get('links', {})
        if links:
            kwargs['query_options'] = links

        recipe = Cls.__new__(Cls, *args, **kwargs)
        recipe.__init__(*args, **kwargs)

        # Like pickle protocol
        if 'state' in entry:
            recipe.__setstate__(entry['state'])

        return recipe

    def get_recipe_object(self, mode):
        """Load recipe object, according to observing mode"""

        if isinstance(mode, ObservingMode):
            key_mode = mode.key
        elif isinstance(mode, str):
            key_mode = mode
        else:
            key_mode = mode

        recipe_entry = self.recipes[key_mode]
        recipe = self._get_base_object(recipe_entry)

        # Init additional members
        recipe.mode = mode
        recipe.instrument = self.instrument

        return recipe

    load_recipe_object = get_recipe_object

    def load_product_object(self, name):
        """Load product object, according to name"""

        product_entry = self.products[name]

        product = self._get_base_object(product_entry)

        return product

    def load_product_class(self, mode):
        """Load recipe object, according to observing mode"""

        product_entry = self.products[mode]

        return self._get_base_class(product_entry)

    def load_product_from_name(self, label):

        short, _ = numina.util.parser.split_type_name(label)
        klass = self.load_product_class(short)
        return klass.from_name(label)

    def depsolve(self):
        """Load all recipes to search for products"""
        # load everything
        requires = {}
        provides = {}
        for mode, r in self.recipes.items():
            robj = self.load_recipe_object(mode)

            for field, vv in robj.requirements().items():
                if vv.type.isproduct():
                    name = vv.type.name()
                    pe = ProductEntry(name, mode, field)
                    requires[name] = pe

            for field, vv in robj.products().items():
                if vv.type.isproduct():
                    name = vv.type.name()
                    pe = ProductEntry(name, mode, field)
                    provides[name] = pe

        return requires, provides

    def who_provides(self, product_label):
        """Return the ProductEntry for some requirement"""

        entry = self._provides_by_p[product_label]
        return entry

    def provides(self, mode_label):
        """Return the ProductEntry for some mode"""

        entry = self._provides_by_r[mode_label]
        return entry

    def _query_recipe(self, thismode, modes):
        if thismode in modes:
            thisnode = modes[thismode]
        else:
            thisnode = numina.core.deptree.DepNode(thismode)
            modes[thismode] = thisnode

        robj = self.load_recipe_object(thismode)
        for field, vv in robj.requirements().items():
            if vv.type.isproduct():

                result = self.who_provides(vv.type.name())
                # get first result
                if result:
                    good = result[0]
                    recurse = False
                    if good.mode not in modes:
                        # Add link
                        recurse = True
                        node = numina.core.deptree.DepNode(good.mode)
                        modes[good.mode] = node

                    node = modes[good.mode]
                    weight = 1
                    if vv.optional:
                        weight = 0
                    # print "link", thismode, "with", good.mode
                    newlink = numina.core.deptree.DepLink(node, weight=weight)
                    thisnode.links.append(newlink)
                    if recurse:
                        self._query_recipe(good.mode, modes)
                else:
                    # No recipe provides this product
                    pass
        return modes

    def query_recipe(self, mode):
        """Recursive query of all calibrations required by a mode"""
        allmodes = self._query_recipe(mode, {})
        return allmodes[mode]


class InstrumentDRP:
    """Description of an Instrument Data Reduction Pipeline

    Parameters
    ==========
       name : str
           Name of the instrument
       configurations : dict of InstrumentConfiguration
       modes : dict of ObservingModes
       pipelines : dict of Pipeline

    """

    def __init__(self, name, configurations, modes, pipelines, products=None, datamodel=None, version='undefined'):
        self.name = name
        self.configurations = configurations
        self.modes = modes
        self.pipelines = pipelines
        if datamodel:
            self.datamodel = datamodel()
        else:
            self.datamodel = numina.datamodel.DataModel()
        self.version = version

    def query_provides(self, product, pipeline='default', search=False):
        """Return the mode that provides a given product"""

        if search:
            return self.search_mode_provides(product)

        pipe = self.pipelines[pipeline]
        try:
            return pipe.who_provides(product)
        except KeyError:
            raise ValueError(f'no mode provides {product}')

    def search_mode_provides(self, product, pipeline='default'):
        """Search the mode that provides a given product"""

        pipeline = self.pipelines[pipeline]
        for obj, mode, field in self.iterate_mode_provides(self.modes, pipeline):
            # extract name from obj
            if obj.name() == product:
                return ProductEntry(obj.name(), mode.key, field)
        else:
            raise ValueError(f'no mode provides {product}')

    def iterate_mode_provides(self, modes, pipeline):
        """Return the mode that provides a given product"""

        for mode_key, mode in modes.items():
            try:
                recipe = pipeline.get_recipe_object(mode_key)
                for key, provide in recipe.products().items():
                    if provide.type.isproduct():
                        yield provide.type, mode, key
            except KeyError:
                warnings.warn(f'Mode {mode_key} has not recipe')

    def configuration_selector(self, obsres):
        warnings.warn("configuration_selector is deprecated, use 'select_configuration' instead",
                      DeprecationWarning, stacklevel=2)
        return self.select_configuration_old(obsres)

    def product_label(self, tipo):
        return tipo.name()

    def select_configuration_old(self, obresult):
        """Select instrument configuration based on OB"""

        logger = logging.getLogger(__name__)
        logger.debug('calling default configuration selector')

        # get first possible image
        ref_frame = obresult.get_sample_frame()
        ref = ref_frame.open()
        extr = self.datamodel.extractor_map['fits']
        if ref:
            # get INSCONF configuration
            result = extr.extract('insconf', ref)
            if result:
                # found the keyword, try to match
                logger.debug('found insconf config uuid=%s', result)
                # Use insconf as uuid key
                if result in self.configurations:
                    return self.configurations[result]
                else:
                    # Additional check for conf.name
                    for conf in self.configurations.values():
                        if conf.name == result:
                            return conf
                    else:
                        raise KeyError(
                            f'insconf {result} does not match any config')

            # If not, try to match by DATE
            date_obs = extr.extract('observation_date', ref)
            for key, conf in self.configurations.items():
                if key == 'default':
                    # skip default
                    continue
                if conf.date_end is not None:
                    upper_t = date_obs < conf.date_end
                else:
                    upper_t = True
                if upper_t and (date_obs >= conf.date_start):
                    logger.debug('found date match, config uuid=%s', key)
                    return conf
        else:
            logger.debug('no match, using default configuration')
            return self.configurations['default']

    def select_configuration(self, obresult):
        return self.select_profile(obresult)

    def select_profile(self, obresult):
        """Select instrument profile based on OB"""

        logger = logging.getLogger(__name__)
        logger.debug('calling default profile selector')
        # check configuration
        insconf = obresult.profile
        if insconf != 'default' and insconf != '00000000-0000-0000-0000-000000000000':
            # Using profile as a UUID
            key = insconf
            date_obs = None
            keyname = 'uuid'
        else:
            # get first possible image
            sample_frame = obresult.get_sample_frame()
            if sample_frame is None:
                key = obresult.instrument
                date_obs = None
                keyname = 'name'
            else:
                return self.select_profile_image(sample_frame.open())
        return key, date_obs, keyname

    def select_profile_image(self, img):
        """Select instrument profile based on FITS"""

        extr = self.datamodel.extractor_map['fits']

        date_obs = extr.extract('observation_date', img)
        key = extr.extract('insconf', img)
        if key is not None:
            keyname = 'uuid'
        else:
            key = extr.extract('instrument', img)
            keyname = 'name'

        return key, date_obs, keyname

    def get_recipe_object(self, mode_name, pipeline_name='default'):
        """Build a recipe object from a given mode name"""
        active_mode = self.modes[mode_name]
        active_pipeline = self.pipelines[pipeline_name]
        recipe = active_pipeline.get_recipe_object(active_mode)
        return recipe


class ProductEntry:
    def __init__(self, name, mode, field, alias=None):
        self.name = name
        self.mode = mode
        self.field = field
        if alias is None:
            split_name = name.split('.')
            self.alias = split_name[-1]
        else:
            self.alias = alias

    def __repr__(self):
        msg = 'ProductEntry(name="{}", mode="{}", field="{}")'.format(
            self.name, self.mode, self.field)
        return msg


class ObservingMode:
    """Observing modes of an Instrument."""

    def __init__(self, instrument=''):
        self.name = ''
        self.key = ''
        self.instrument = instrument
        self.summary = ''
        self.description = ''
        self.tagger = None
        self.validator = None
        self.build_ob_options = None
        self.rawimage = None

    def validate(self, obsres):
        return True

    def build_ob(self, partial_ob, backend, options=None):

        mod = options or self.build_ob_options

        if isinstance(mod, numina.core.query.ResultOf):
            result_type = mod.result_type
            name = 'relative_result'
            val = backend.search_result_relative(
                name, result_type, partial_ob, result_desc=mod)
            for r in val:
                partial_ob.results[r.id] = r.content

        return partial_ob

    def tag_ob(self, partial):
        if self.tagger is not None:
            warnings.warn("per mode taggers are deprecated, recipe requirements provide al required information",
                          DeprecationWarning, stacklevel=2)
            partial.tags = self.tagger(partial)
        return partial

    def __repr__(self):
        return f"ObservingMode(name={self.name})"




import pkgutil

import pytest

import numina
import numina.drps
import numina.drps.drpbase
import numina.core.pipelineload as pload

from ..cli import main


drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')

expecte0 = [""]


expecte1 = ["Instrument: TEST1",
            " version is '1'",
            " has profile 'Old configuration' uuid=6ad5dc90-6b15-43b7-abb5-b07340e19f41",
            " has profile 'Default configuration' uuid=225fcaf2-7f6f-49cc-972a-70fd0aee8e96",
            " has datamodel 'numina.datamodel.DataModel'",
            " has pipeline 'default', version 1",
            ""
            ]


expecte2 = ["Instrument: TEST2",
            f" version is '{numina.__version__}'",
            " has datamodel 'numina.datamodel.DataModel'",
            " has profile 'Default configuration' uuid=9c21b315-9231-4fe0-a276-5043b064a3a8",
            " has pipeline 'default', version 1",
            "Instrument: TEST1",
            " version is '1'",
            " has profile 'Old configuration' uuid=6ad5dc90-6b15-43b7-abb5-b07340e19f41",
            " has profile 'Default configuration' uuid=225fcaf2-7f6f-49cc-972a-70fd0aee8e96",
            " has datamodel 'numina.datamodel.DataModel'",
            " has pipeline 'default', version 1",
            ""
            ]


def mockreturn0():
    return numina.drps.drpbase.DrpGeneric()


def mockreturn1():
    drps = {}
    drp1 = pload.drp_load_data('numina', drpdata1)
    drps[drp1.name] = drp1
    return numina.drps.drpbase.DrpGeneric(drps)


def mockreturn2():
    drps = {}
    drp1 = pload.drp_load_data('numina', drpdata1)
    drp2 = pload.drp_load_data('numina', drpdata2)
    drps[drp1.name] = drp1
    drps[drp2.name] = drp2
    return numina.drps.drpbase.DrpGeneric(drps)


@pytest.mark.parametrize("drpsfunc, expected", [
    (mockreturn0, expecte0),
    (mockreturn1, expecte1),
    (mockreturn2, expecte2)
])
def test_show_instrument(capsys, monkeypatch, drpsfunc, expected):
    """Test that no instruments are shown"""

    monkeypatch.setattr(numina.drps, "get_system_drps", drpsfunc)

    main(['show-instruments'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


@pytest.mark.parametrize("drpsfunc", [mockreturn0, mockreturn1, mockreturn2])
def test_show_no_instruments_i(capsys, monkeypatch, drpsfunc):
    """Test that two instruments are shown"""

    monkeypatch.setattr(numina.drps, "get_system_drps", drpsfunc)

    expected = ["No instrument named: TEST3",
                ""
                ]

    main(['show-instruments', '-i', 'TEST3'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected



import pytest

from numina.drps.tests.pattridrp.instrument.components.robot import RobotPos, Robot


def test_robot1(pattri_ins):
    dev = pattri_ins.get_device("robot")
    assert isinstance(dev, RobotPos)


def test_robot2(pattri_ins):
    base_dev = pattri_ins.get_device("robot")
    for dev in base_dev.children.values():
        assert isinstance(dev, Robot)


def test_robot3(pattri_ins):
    base_dev = pattri_ins.get_device("robot")
    for name, dev in base_dev.children.items():
        assert name.startswith("arm_")
        props = dev.get_property_names()
        assert "active" in props
        assert "angle" in props


def test_robot4(pattri_ins):
    base_dev = pattri_ins.get_device("robot")
    assert len(base_dev.children) == 7


def test_robot_hdr(pattri_ins, pattri_header):
    dev = pattri_ins.get_device("robot")
    hdr = pattri_header
    dev.configure_with_header(hdr)
    assert pattri_ins.get_device("robot.arm_1").angle == -0.3
    assert pattri_ins.get_device("robot.arm_2").active is True
    assert pattri_ins.get_device("robot.arm_3").active is False
    assert pattri_ins.get_device("robot.arm_3").angle == 0.4


def test_robot_hdr2(pattri_ins, pattri_header2):
    dev = pattri_ins.get_device("robot")
    dev.configure_with_header(pattri_header2)
    assert pattri_ins.get_device("robot.arm_1").angle == -0.3
    assert pattri_ins.get_device("robot.arm_2").active is False
    assert pattri_ins.get_device("robot.arm_3").active is False
    assert pattri_ins.get_device("robot.arm_3").angle == 0.4


def test_robot_active1():
    """If robot is not active, angle cant be changed"""
    dev = Robot("Robo", {}, 0, active=False)
    with pytest.warns(UserWarning):
        dev.angle = 2.4
    assert dev.angle == 0


def test_robot_active2():
    """If robot is not active, it can be configured"""
    dev = Robot("Robo", {}, 0, active=False)
    with dev.state.managed_configure():
        dev.angle = 2.4
    assert dev.angle == 2.4


def test_instrument2(pattri_ins):
    dev = pattri_ins.get_device("robot")
    for sub in dev.children:
        sub_dev = dev.children[sub]
        assert isinstance(sub_dev, Robot)
        assert sub_dev.name == sub

    assert isinstance(dev, RobotPos)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Correct cosmic rays applying previously computed masks."""
import logging

from astropy.table import Table
import numpy as np
import numpy.ma as ma
from rich.logging import RichHandler

from .valid_parameters import VALID_COMBINATIONS


def apply_crmasks(list_arrays, hdul_masks=None, combination=None, use_lamedian=None,
                  dtype=np.float32, apply_flux_factor=None, bias=None):
    """
    Correct cosmic rays applying previously computed masks.

    The input arrays are bias subtracted, and optionally re-scaled by
    a flux factor read from the header of `hdul_masks`. Then, they are
    combined using the specified combination method.

    Parameters
    ----------
    list_arrays : list of 2D arrays
        The input arrays to be combined.
    hdul_masks : HDUList
        The HDUList containing the mask arrays for cosmic ray removal.
        The masks and auxiliary data should be in the following extensions:
        - 'MEDIANCR' mask for the median combination
        - 'MEANCRT' mask for the mean combination
        - 'CRMASK1', 'CRMASK2', etc. masks for the mean combination with
          individual masks
        - 'LAMEDIAN' (optional) image with the lacosmic-corrected median array
        - 'RPMEDIAN' (optional) pixels to be replaced by the local median value
          around them. Stored as a binary table with four columns:
          'X_pixel', 'Y_pixel', 'X_width', 'Y_width'. The (X, Y) coordinates
          are given following the FITS convention (starting at 1).
    combination : str
        The type of combination to apply. There are three options:
        - 'mediancr', the median combination is applied, and masked pixels
        (those equal to 1 in extension 'MEDIANCR' of `hdul_masks`) are
        replaced by the minimum value of the corresponding pixel in the
        input arrays.
        - 'meancrt', the mean combination is applied, and masked pixels
        (those equal to 1 in extension 'MEANCRT' of `hdul_masks`) are
        replaced by the mediancr value.
        - 'meancr', the mean combination is applied making use of
        the individual mask of each image (extensions 'CRMASK1', 'CRMASK2',
        etc. in `hdul_masks`). Those pixels that are masked in all the individual
        images are replaced by the minimum value of the corresponding pixel
        in the input arrays.
    use_lamedian : bool, optional
        If True, and if the extension 'LAMEDIAN' is present in `hdul_masks`,
        the lacosmic-corrected median array is used instead of the minimum
        value at each pixel. This affects differently depending on the
        combination method:
        - 'mediancr': all the masked pixels in the mask MEDIANCR are replaced.
        - 'meancrt': only the pixels coincident in masks MEANCRT and MEDIANCR;
          the rest of the pixels flagged in the mask MEANCRT are replaced by
          the value obtained when the combination method is 'mediancr'.
        - 'meancr': only the pixels flagged in all the individual exposures
          (i.e., those flagged simulatenously in all the CRMASKi masks);
          the rest of the pixels flagged in any of the `CRMASK1`, `CRMASK2`, etc.
          masks are replaced by the corresponding masked mean.
        Default is False.
    apply_flux_factor : bool, optional
        If True, the flux factor is applied to the input arrays before
        combining them. The flux factor is read from the header of the
        `hdul_masks` (keywords 'FLUXF1', 'FLUXF2', etc.). Default is True.
    bias : float or 2D array, optional
        The bias level to be subtracted from the input arrays. If a float is
        provided, it is assumed to be constant for all pixels. If a 2D array
        is provided, it must have the same shape as the input arrays. If None,
        the bias is assumed to be zero for all pixels. Default is None.
    dtype : data-type, optional
        The desired data type for the output arrays (default is np.float32).

    Returns
    -------
    combined2d: 2D array
        The combined bias-subtracted array with masked pixels replaced
        accordingly depending on the combination method.
    variance2d : 2D array
        The variance of the input arrays along the first axis.
    map2d : 2D array
        The number of input pixels used to compute the median at each pixel.
    """

    _logger = logging.getLogger(__name__)
    rich_configured = any(isinstance(handler, RichHandler) for handler in _logger.handlers)

    # Check that the input is a list
    if not isinstance(list_arrays, list):
        raise TypeError("Input must be a list of arrays.")

    # Check that the combination method is valid
    if combination not in VALID_COMBINATIONS:
        raise ValueError(f"Combination: {combination} must be one of {VALID_COMBINATIONS}.")

    # If use_lamedian is True, check that the extension 'LAMEDIAN' is present
    if use_lamedian:
        if 'LAMEDIAN' not in hdul_masks:
            raise ValueError("use_lamedian is True, but extension 'LAMEDIAN' is not present in hdul_masks.")

    # Check that the list contains numpy 2D arrays
    if not all(isinstance(array, np.ndarray) and array.ndim == 2 for array in list_arrays):
        raise ValueError("All elements in the list must be 2D numpy arrays.")

    # Check that the list contains at least 3 arrays
    num_images = len(list_arrays)
    if num_images < 3:
        raise ValueError("At least 3 images are required for a useful combination.")

    # Check that all arrays have the same shape
    for i, array in enumerate(list_arrays):
        if array.shape != list_arrays[0].shape:
            raise ValueError(f"Array {i} has a different shape than the first array.")
    naxis2, naxis1 = list_arrays[0].shape

    # Log the number of input arrays and their shapes
    _logger.info("number of input arrays: %d", len(list_arrays))
    for i, array in enumerate(list_arrays):
        _logger.info("array %d shape: %s, dtype: %s", i, array.shape, array.dtype)

    # Define the bias
    if bias is None:
        bias = np.zeros((naxis2, naxis1), dtype=float)
        _logger.info("bias not defined, assuming bias=0.0 for all pixels.")
    elif isinstance(bias, (float, int)):
        bias = np.full((naxis2, naxis1), bias, dtype=float)
        _logger.info("bias defined as a constant value: %f", bias[0, 0])
    elif isinstance(bias, np.ndarray):
        if bias.shape != (naxis2, naxis1):
            raise ValueError(f"bias must have the same shape as the input arrays ({naxis2=}, {naxis1=}).")
        _logger.info("bias defined as a 2D array with shape: %s", bias.shape)
    else:
        raise TypeError(f"Invalid type for bias: {type(bias)}. Must be float, int, or numpy array.")

    # Read the flux factor from the masks
    if apply_flux_factor is None:
        raise ValueError("apply_flux_factor must be specified as True or False.")
    _logger.info("apply_flux_factor: %s", apply_flux_factor)
    flux_factor = []
    for i in range(num_images):
        flux_factor.append(hdul_masks[0].header[f'FLUXF{i+1}'])
    flux_factor = np.array(flux_factor, dtype=float)
    if apply_flux_factor:
        _logger.info("flux factor values: %s will be employed", str(flux_factor))
    else:
        _logger.info("flux factor values: %s will NOT be employed", str(flux_factor))
        flux_factor = np.ones(num_images, dtype=float)
        _logger.info("flux factor values set to %s", str(flux_factor))

    # Check use_lamedian
    if use_lamedian is None:
        raise ValueError("use_lamedian must be specified as True or False.")
    _logger.info("use_lamedian: %s", use_lamedian)

    # Convert the list of arrays to a 3D numpy array
    shape3d = (num_images, naxis2, naxis1)
    image3d = np.zeros(shape3d, dtype=dtype)
    _logger.info("applying bias and flux factors to input arrays...")
    for i, array in enumerate(list_arrays):
        image3d[i] = (array.astype(dtype) - bias) / flux_factor[i]

    # Compute minimum and median along the first axis of image3d
    min2d_rescaled = np.min(image3d, axis=0)
    median2d = np.median(image3d, axis=0)

    # Apply the requested combination method
    if rich_configured:
        rlabel_combination = f"[bold][magenta]{combination}[/magenta][/bold]"
    else:
        rlabel_combination = f"{combination}"
    _logger.info("applying combination method: %s", rlabel_combination)
    _logger.info("using crmasks in %s", hdul_masks[0].header['UUID'])
    if combination in ['mediancr', 'meancrt']:
        # Define the mask_mediancr
        mask_mediancr = hdul_masks['MEDIANCR'].data.astype(bool)
        _logger.info("applying mask MEDIANCR: %d masked pixels", np.sum(mask_mediancr))
        # Replace the masked pixels with the minimum value
        # of the corresponding pixel in the input arrays
        median2d_corrected = median2d.copy()
        if use_lamedian:
            median2d_corrected[mask_mediancr] = hdul_masks['LAMEDIAN'].data[mask_mediancr]
        else:
            median2d_corrected[mask_mediancr] = min2d_rescaled[mask_mediancr]

    if combination == 'mediancr':
        combined2d = median2d_corrected
        # Define the variance and map arrays
        variance2d = np.var(image3d, axis=0, ddof=1)
        variance2d[mask_mediancr] = 0.0  # Set variance to 0 for the masked pixels
        map2d = np.ones((naxis2, naxis1), dtype=int) * num_images
        map2d[mask_mediancr] = 1  # Set the map to 1 for the masked pixels
    elif combination == 'meancrt':
        # Define the mask_meancr
        mask_meancrt = hdul_masks['MEANCRT'].data.astype(bool)
        _logger.info("applying mask MEANCRT: %d masked pixels", np.sum(mask_meancrt))
        mean2d = np.mean(image3d, axis=0)
        # Replace the masked pixels in mean2d with the median2d_corrected value
        mean2d_corrected = mean2d.copy()
        mean2d_corrected[mask_meancrt] = median2d_corrected[mask_meancrt]
        combined2d = mean2d_corrected
        # Define the variance and map arrays
        variance2d = np.var(image3d, axis=0, ddof=1)
        variance2d[mask_meancrt] = 0.0  # Set variance to 0 for the masked pixels
        map2d = np.ones((naxis2, naxis1), dtype=int) * num_images
        map2d[mask_meancrt] = 1  # Set the map to 1 for the masked pixels
    elif combination == 'meancr':
        image3d_masked = ma.array(
            np.zeros(shape3d, dtype=dtype),
            mask=np.full(shape3d, fill_value=True, dtype=bool)
        )
        # Loop through each image and apply the corresponding mask
        total_mask = np.zeros((naxis2, naxis1), dtype=int)
        for i in range(num_images):
            image3d_masked[i, :, :] = image3d[i, :, :]
            mask = hdul_masks[f'CRMASK{i+1}'].data
            _logger.info("applying mask %s: %d masked pixels", f'CRMASK{i+1}', np.sum(mask))
            total_mask += mask.astype(int)
            image3d_masked[i, :, :].mask = mask.astype(bool)
        # Compute the mean of the masked 3D array
        combined2d = ma.mean(image3d_masked, axis=0).data
        # Replace pixels without data with the minimum value
        mask_nodata = total_mask == num_images
        if np.any(mask_nodata):
            if use_lamedian:
                _logger.info("replacing %d pixels without data by the LAMEDIAN value", np.sum(mask_nodata))
                combined2d[mask_nodata] = hdul_masks['LAMEDIAN'].data[mask_nodata]
            else:
                _logger.info("replacing %d pixels without data by the minimum value", np.sum(mask_nodata))
                combined2d[mask_nodata] = min2d_rescaled[mask_nodata]
        else:
            _logger.info("no pixels without data found, no replacement needed")
        # Define the variance and map arrays
        variance2d = ma.var(image3d_masked, axis=0, ddof=1).data
        map2d = np.ones((naxis2, naxis1), dtype=int) * num_images - total_mask
    else:
        raise ValueError(f"Invalid combination method: {combination}. "
                         f"Valid options are {VALID_COMBINATIONS}.")

    # Replaced pixels defined in the RPMEDIAN extension of hdul_masks by
    # the local median around them (excluding the considered pixel)
    if 'RPMEDIAN' in hdul_masks:
        _logger.info("replacing pixels defined in RPMEDIAN by the local median around them")
        table_rpmedian = Table(hdul_masks['RPMEDIAN'].data)
        for row in table_rpmedian:
            xpixel = row['X_pixel'] - 1  # Convert from FITS to zero-based index
            ypixel = row['Y_pixel'] - 1  # Convert from FITS to zero-based index
            xwidth = row['X_width']
            ywidth = row['Y_width']
            # Define the box around the pixel, ensuring it is within image bounds
            x1 = max(xpixel - xwidth // 2, 0)
            x2 = min(xpixel + xwidth // 2 + 1, naxis1)
            y1 = max(ypixel - ywidth // 2, 0)
            y2 = min(ypixel + ywidth // 2 + 1, naxis2)
            # Extract the local box from the combined2d array
            local_box = combined2d[y1:y2, x1:x2]
            # Create a mask for the local box that excludes the central pixel
            local_mask = np.ones(local_box.shape, dtype=bool)
            if y1 <= ypixel < y2 and x1 <= xpixel < x2:
                local_mask[ypixel - y1, xpixel - x1] = False
            # Compute the median of the unmasked pixels in the local box
            if np.any(local_mask):
                local_median = np.median(local_box[local_mask])
                _logger.info(f"- pixel (x={xpixel+1}, y={ypixel+1}): "
                             f"old value={combined2d[ypixel, xpixel]:.2f}, new value={local_median:.2f}")
                combined2d[ypixel, xpixel] = local_median
            else:
                _logger.warning(f"cannot compute local median for pixel (x={xpixel+1}, y={ypixel+1}) "
                                "as no surrounding pixels are available")

    return combined2d.astype(dtype), variance2d.astype(dtype), map2d



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Corrections of wavelength calibration"""

import logging
import warnings

import astropy.wcs
from astropy.coordinates import SkyCoord, EarthLocation
import astropy.time
import astropy.constants as cons

_logger = logging.getLogger(__name__)


def header_add_barycentric_correction(hdr, key='B', out=None):
    """Add WCS keywords with barycentric correction

    Raises
    ------
    KeyError
        If a required keyword is missing
    TypeError
        If the header does not contain a spectral axis
    """

    # Header must have DATE-OBS
    if 'DATE-OBS' not in hdr:
        raise KeyError("Keyword 'DATE-OBS' not found.")
    # Header must contain a primary WCS
    # Header must contain RADEG and DECDEG

    if 'OBSGEO-X' not in hdr:
        warnings.warn(
            'OBSGEO- keywords not defined, using default values for GTC', RuntimeWarning)
        # Geocentric coordinates of GTC
        hdr['OBSGEO-X'] = 5327285.0921
        hdr['OBSGEO-Y'] = -1718777.1125
        hdr['OBSGEO-Z'] = 3051786.7327

    # Get main WCS
    wcs0 = astropy.wcs.WCS(hdr)
    if wcs0.wcs.spec == -1:
        # We don't have a spec axis
        raise TypeError('Header does not contain spectral axis')
    gtc = EarthLocation.from_geocentric(
        wcs0.wcs.obsgeo[0], wcs0.wcs.obsgeo[1], wcs0.wcs.obsgeo[2], unit='m')
    date_obs = astropy.time.Time(wcs0.wcs.dateobs, format='fits')
    # if frame='fk5', we need to pass the epoch and equinox
    sc = SkyCoord(ra=hdr['RADEG'], dec=hdr['DECDEG'], unit='deg')
    rv = sc.radial_velocity_correction(obstime=date_obs, location=gtc)
    factor = (1 + rv / cons.c).to('').value

    if out is None:
        out = hdr

    out[f'WCSNAME{key}'] = 'Barycentric correction'
    # out['CNAME1{}'.format(key)] = 'AxisV'
    out[f'CTYPE1{key}'] = hdr['CTYPE1']
    out[f'CRPIX1{key}'] = hdr['CRPIX1']
    out[f'CRVAL1{key}'] = hdr['CRVAL1'] * factor
    out[f'CDELT1{key}'] = hdr['CDELT1'] * factor
    out[f'CUNIT1{key}'] = hdr['CUNIT1']

    for keyword in ['CRPIX2', 'CRVAL2', 'CDELT2']:
        out[f'{keyword}{key}'] = hdr[f'{keyword}']

    out[f'VELOSYS{key}'] = rv.to('m / s').value
    out[f'SPECSYS{key}'] = 'BARYCENT'
    out[f'SSYSOBS{key}'] = 'TOPOCENT'
    return out



import numpy

"""Create simple objects for testing"""


def create_simple_hdul(shape=(1, 1), value=0.0):
    """Create a simple image for testing"""
    import astropy.io.fits as fits

    data = value * numpy.ones(shape, dtype='float32')
    header = fits.Header()
    header['DATE-OBS'] = '2001-09-12T22:45:12.343'
    simple_img = fits.HDUList([fits.PrimaryHDU(data=data, header=header)])
    return simple_img


def create_simple_frame(shape=(1, 1), value=0.0):
    """Create a simple image frame for testing"""
    from numina.types.frame import DataFrame

    simple_img = create_simple_hdul(shape, value=value)
    simple_frame = DataFrame(frame=simple_img)
    return simple_frame


def create_simple_structured():
    """Create a simple structred object for testing"""
    from numina.types.structured import BaseStructuredCalibration
    import numina.types.qc as qc

    obj = BaseStructuredCalibration()
    obj.quality_control = qc.QC.BAD
    return obj




import pytest


from numina.util.parser import parse_arg_line
from numina.util.parser import split_type_name


def test_split_type_name():
    label = 'TraceMap(a=1, b="2")'

    pre, post = split_type_name(label)

    assert pre == 'TraceMap'
    assert post == 'a=1, b="2"'


def test_parse_arg():

    assert parse_arg_line("a=1, b='2', c=True") == {
        "a": 1, "b": "2", "c": True}


def test_parse_arg_empty():

    assert parse_arg_line("   ") == {}


def test_parse_arg_malformed():

    with pytest.raises(ValueError):
        parse_arg_line("abcbcb")



#
# Copyright 2016-2035 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
import typing

if typing.TYPE_CHECKING:
    from .configorigin import ElementOrigin

from .state import State


class DeviceBase:
    def __init__(
        self,
        name: str,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
    ):
        self.name = name
        self.parent = None
        self.origin = None
        self.uuid = None
        self.children: dict[str, "DeviceBase"] = {}
        self.set_origin(origin)
        self.set_parent(parent)
        self.state = State()
        # FIXME: this can be removed
        self.is_configured = False

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
        properties=None,
        setup=None,
    ) -> "DeviceBase":
        obj = cls.__new__(cls)
        obj.__init__(comp_id, origin, parent)
        return obj

    def set_origin(self, origin: "ElementOrigin") -> None:
        self.origin = origin
        if self.origin is not None:
            self.uuid = self.origin.uuid

    def set_parent(self, new_parent: "DeviceBase") -> None:
        if self.parent:
            del self.parent.children[self.name]
        if new_parent:
            if self.name in new_parent.children:
                raise ValueError(f"{self.name} already registered with new_parent")
            else:
                new_parent.children[self.name] = self
                self.parent = new_parent

    def get_device(self, path: str) -> "DeviceBase":
        vals = path.split(".")
        return self.get_device_seq(vals)

    def get_device_seq(self, path: list[str]) -> "DeviceBase":

        if len(path) == 0:
            return self

        base = path[0]
        rpath = path[1:]
        if base == self.name:
            # Skip first term if its myself
            return self.get_device_seq(rpath)

        comp = self.children[base]
        if rpath:
            return comp.get_device_seq(rpath)
        else:
            return comp

    @staticmethod
    def split_path(path: str) -> tuple[list[str], str]:
        keys = path.split(".")
        device_path = keys[:-1]
        property_name = keys[-1]
        return device_path, property_name

    def get_property(self, path: str):
        device_path, prop_name = self.split_path(path)
        dev = self.get_device_seq(device_path)
        val = getattr(dev, prop_name)
        return val

    def get_value(self, path: str, **state):
        # Ignoring state
        return self.get_property(path)

    def configure_me(self, info: dict) -> None:
        with self.state.managed_configure():
            for key in info:
                setattr(self, key, info[key])

    def configure(self, info: dict) -> None:
        for key, value in info.items():
            node = self.get_device(key)
            if node:
                node.configure_me(value)

    def configure_me_with_header(self, hdr) -> None:
        with self.state.managed_configure():
            # To be implemented...
            pass

    def configure_with_header(self, hdr) -> None:
        self.configure_me_with_header(hdr)
        for ch in self.children.values():
            ch.configure_with_header(hdr)

    def configure_me_with_image(self, image) -> None:
        # To be implemented...
        self.configure_me_with_header(image[0].header)

    def configure_with_image(self, image):
        self.configure_me_with_image(image)
        for ch in self.children.values():
            ch.configure_with_image(image)

    def depends_on(self):
        """Compute the dependencies for me and my children"""
        total_depends = set([])

        total_depends.update(self.my_depends())

        for ch in self.children.values():
            ch_depends = ch.depends_on()
            total_depends.update(ch_depends)
        return total_depends

    def my_depends(self) -> set[str]:
        return set([])

    def get_property_names(self) -> set[str]:
        return set([])

    def init_config_info(self) -> dict[str, typing.Any]:
        return dict()

    def config_info(self):
        return visit(self)

    def end_config_info(self, result) -> dict[str, typing.Any]:
        if self.children:
            result["children"] = list(self.children.keys())
        return result

    def get_properties(self, init=True) -> dict[str, typing.Any]:
        if init:
            result = self.init_config_info()
        else:
            result = dict()

        for key in self.get_property_names():
            result[key] = getattr(self, key)
        return result


def visit(node: DeviceBase, root: str = "", result: dict | None = None, sep: str = "."):
    """Visit recursively all subdevices of node and return their configurations"""

    result = {} if result is None else result

    if root != "":
        node_name = root + sep + node.name
    else:
        node_name = node.name

    result[node_name] = node.get_properties()
    sub_meta = result
    for child in node.children.values():
        visit(child, root=node_name, result=sub_meta)
    return result



#
# Copyright 2019-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import os
import shutil
import tempfile

import pytest

import numina.util.context as ctx
from numina.util.jsonencoder import ExtEncoder


class ResultCompPlugin(object):
    """Plugin to compare the results fo reductions"""

    def __init__(self, config, reference_dir=None, generate_dir=None):
        self.config = config
        self.reference_dir = reference_dir
        self.generate_dir = generate_dir

    def pytest_runtest_setup(self, item):
        import functools
        compare = item.get_closest_marker('result_compare')

        if compare is None:
            return

        original = item.function

        atol = compare.kwargs.get('atol', 0.)
        rtol = compare.kwargs.get('rtol', 1e-7)

        @functools.wraps(item.function)
        def item_function_wrapper(*args, **kwargs):

            reference_dir = compare.kwargs.get('reference_dir', None)
            if reference_dir is None:
                if self.reference_dir is None:
                    reference_dir = os.path.join(
                        os.path.dirname(item.fspath.strpath), 'reference')
                else:
                    reference_dir = self.reference_dir
            else:
                if not reference_dir.startswith(('http://', 'https://')):
                    reference_dir = os.path.join(os.path.dirname(
                        item.fspath.strpath), reference_dir)

            baseline_remote = reference_dir.startswith('http')

            # Run test and get result object
            import inspect
            if inspect.ismethod(original):  # method
                result = original(*args[1:], **kwargs)
            else:  # function
                result = original(*args, **kwargs)

            # Task or result...
            destination = compare.kwargs.get('destination', None)
            if destination is None:
                destination = item.name
                destination = destination.replace('[', '_').replace(']', '_')

            if self.generate_dir is None:

                # Save the result
                result_dir = tempfile.mkdtemp()
                with ctx.working_directory(result_dir):
                    manifest = generate_manifest(result)

                    import json
                    with open('result.json', 'w') as fd:
                        json.dump(manifest, fd, indent=2, cls=ExtEncoder)

                if baseline_remote:
                    # baseline_file_ref = _download_file(reference_dir + filename)
                    raise NotImplementedError
                else:
                    baseline_file_ref = os.path.abspath(
                        os.path.join(os.path.dirname(
                            item.fspath.strpath), reference_dir, destination)
                    )

                if not os.path.exists(baseline_file_ref):
                    exmsg = "File not found for comparison test\n" \
                            "Generated file:\t{test}\n" \
                            "This is expected for new tests."
                    raise Exception(exmsg.format(test=destination))

                # Compare my result with something else
                identical, msg = compare_result_dirs(
                    baseline_file_ref, result_dir, atol=atol, rtol=rtol)

                if identical:
                    shutil.rmtree(result_dir)
                else:
                    raise Exception(msg)
            else:
                # Generating
                generate_detail = os.path.join(self.generate_dir, destination)
                if not os.path.exists(generate_detail):
                    os.makedirs(generate_detail)

                with ctx.working_directory(generate_detail):

                    manifest = generate_manifest(result)

                    import json
                    with open('result.json', 'w') as fd:
                        json.dump(manifest, fd)

                # Write something in destination...
                pytest.skip("skipping test, since generating data")

        # Override with wrapped version
        if item.cls is not None:
            setattr(item.cls, item.function.__name__, item_function_wrapper)
        else:
            item.obj = item_function_wrapper


def compare_eq_sequence(left, right):
    explanation = []
    for i in range(min(len(left), len(right))):
        if left[i] != right[i]:
            explanation += [f"At index {i} diff: {left[i]!r} != {right[i]!r}"]
            break
    if len(left) > len(right):
        explanation += [
            f"Left contains more items, first extra item: {left[len(right)]}"
        ]
    elif len(left) < len(right):
        explanation += [
            f"Right contains more items, first extra item: {right[len(left)]}"
        ]
    return explanation


def compare_result_dirs(resdir1, resdir2, atol=0.0, rtol=1e-7):
    # First is reference, second is generated
    # print('compare contents of', resdir1, resdir2)
    # Compare dir contents
    cont1 = sorted(os.listdir(resdir1))
    cont2 = sorted(os.listdir(resdir2))

    # as assert cont1 == cont2
    if cont1 != cont2:
        exp = compare_eq_sequence(cont1, cont2)
        return False, exp

    for fname in cont1:
        fname1 = os.path.join(resdir1, fname)
        fname2 = os.path.join(resdir2, fname)
        part, ext = os.path.splitext(fname)
        # Check extensions
        if ext == '.fits':
            from astropy.io.fits.diff import FITSDiff
            ignore_keywords = [
                'HISTORY', 'UUID', 'NUMXVER'
            ]
            diff = FITSDiff(
                fname1, fname2, rtol=rtol, atol=atol,
                ignore_keywords=ignore_keywords
            )
            return diff.identical, diff.report()
        elif ext == '.json':
            print('json comparator, not implemented')
        else:
            print('ignoring ', ext, 'file')
    return True, "ok"


def generate_manifest(recipe_result):
    import numina.store

    saveres = dict(values={})
    saveres_v = saveres['values']
    for key, prod in recipe_result.stored().items():
        val = getattr(recipe_result, key)
        saveres_v[key] = numina.store.dump(prod.type, val, prod.dest)

    saveres['qc'] = recipe_result.qc.name
    saveres['uuid'] = str(recipe_result.uuid)

    return saveres




import astropy.io.fits as fits
import numpy

from numina.frame.utils import copy_img


def test_copy_img():

    hdu1 = fits.PrimaryHDU(data=[1, 2, 3])

    hdr2 = fits.Header({'A': 1})
    hdu2 = fits.ImageHDU(data=[3.0, 4.3, 999], header=hdr2)

    hdu3 = fits.ImageHDU(data=[[1, 2, 3, 4, 5]])

    hdul = fits.HDUList([hdu1, hdu2, hdu3])

    hdul_copy = copy_img(hdul)

    assert hdul is not hdul_copy
    assert len(hdul_copy) == len(hdul_copy)
    for i in range(3):
        assert numpy.allclose(hdul[i].data, hdul_copy[i].data)
        assert hdul[i].header == hdul_copy[i].header




import numpy as np
from numpy.testing import assert_allclose
from ..numsplines import AdaptiveLSQUnivariateSpline


def test_adaptive_sqlunivariatespline():
    npoints = 300
    np.random.seed(1234)
    xdata = np.linspace(-2*np.pi, 2*np.pi, npoints)
    ydata = np.sin(xdata) + 0.3 * np.random.randn(npoints)

    # adaptive splines
    spl = AdaptiveLSQUnivariateSpline(xdata, ydata, 2)
    xknots = spl.get_knots()
    yknots = spl(xknots)

    assert_allclose(
        spl.get_coeffs(),
        [-0.2450535, 3.15287065, -3.6363213, 3.60549501,
         -2.96073338,  0.18723241], rtol=1e-6)
    assert_allclose(spl.get_residual(), 26.538044076506353)
    assert_allclose(xknots,
                    [-6.28318531, -1.34353445, 1.30888191, 6.28318531],
                    rtol=1e-6)
    assert_allclose(yknots,
                    [-0.2450535, -0.95552754, 0.94163857, 0.18723241],
                    rtol=1e-6)

    # fixed knots
    spl = AdaptiveLSQUnivariateSpline(xdata, ydata, [-4, 0, 4], adaptive=False)
    result = spl.get_result()
    assert result is None
    xknots = spl.get_knots()
    yknots = spl(xknots)
    assert_allclose(
        spl.get_coeffs(),
        [-0.65286453, 2.38128699, -1.26391753, -0.05672272,
         1.37701644, -2.33129953, 0.61925078],
        rtol=1e-6
    )
    assert_allclose(spl.get_residual(), 59.696265273407434)
    assert_allclose(
        xknots, [-6.28318531, -4., 0., 4., 6.28318531], rtol=1e-6
    )
    assert_allclose(yknots, [-0.65286453, 0.31082669, -0.01266159,
                             -0.24158386, 0.61925078], rtol=1e-6)




import pytest

import numina.core
import numina.core.query
import numina.dal.stored
import numina.exceptions
import numina.types.datatype

from ..dataholders import Requirement


class Dal(object):
    def search_parameter(self, name, type_, obsres, options=None):
        if name == 'req_null':
            return numina.dal.stored.StoredParameter(content=None)
        elif name == 'req_int':
            return numina.dal.stored.StoredParameter(content=2)
        else:
            raise numina.exceptions.NoResultFound('value not found')


def test_query1():

    req = Requirement(rtype=None, description="", destination='req_null')

    dal = Dal()

    obres = numina.core.ObservationResult()

    result = req.query(dal, obres)
    assert result is None


def test_query11():

    req = Requirement(rtype=numina.types.datatype.PlainPythonType(
        1), description="", destination='req_int')

    dal = Dal()

    obres = numina.core.ObservationResult()

    result = req.query(dal, obres)
    assert result == 2
    # assert False


def test_dest_not_set():

    req = Requirement(rtype=None, description="")

    dal = Dal()

    obres = numina.core.ObservationResult()

    with pytest.raises(ValueError):
        req.query(dal, obres)



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary script to perform binary image arithmetic.

The computation is performed casting to np.float32.
"""

import argparse
import sys

from astropy.io import fits
import numpy as np
import pathlib

from numina.array.display.ximshow import ximshow_file, ximshow


def extnum_from_extname(file, extname):
    """Get the extension number from the extension name in a FITS file.

    Parameters
    ----------
    file : file object
        FITS file.
    extname : str
        Extension name.

    Returns
    -------
    int
        Extension number.
    """
    with fits.open(file) as hdulist:
        for i, hdu in enumerate(hdulist):
            if hdu.name.lower() == extname.lower():
                return i

    raise ValueError(f"Extension '{extname}' not found in {file.name}.")


def compute_operation(
    file1,
    file2,
    operation,
    extname1,
    extname2,
    output,
    overwrite=True,
    dtype='float32',
    display='none',
    args_z1z2=None,
    args_bbox=None,
    args_keystitle=None,
    args_geometry=None
):
    """Compute output = file1 operation file2.

    Parameters
    ----------
    file1 : file object
        First FITS file.
    file2 : file object or float
        Second FITS file or float number.
    operation : string
        Mathematical operation.
    extname1 : str
        Extension name of the first FITS file.
    extname2 : str
        Extension name of the second FITS file.
    output : file object
        Output FITS file.
    overwrite : bool
        If True, the output file can be overwritten.
    dtype : str
        Data type of the output image. Default is 'float32'.
    display : string
        Character string indication whether the images are displayed.
        Valid values are 'all', 'result' and 'none' (default).
    args_z1z2 : string or None
        String providing the image cuts tuple: z1, z2, minmax or None.
    args_bbox : string or None
        String providing the bounding box tuple: nc1, nc2, ns1, ns2.
    args_keystitle : string or None
        Tuple of FITS keywords.format: key1,key2,...,keyn.format
    args_geometry : string or None
        Tuple x,y,dx,dy to define the Qt backend geometry.

    """

    # read first FITS file
    with fits.open(file1) as hdulist:
        image_header1 = hdulist[extname1].header
        image1 = hdulist[extname1].data.astype(dtype)
    naxis = image_header1['naxis']
    naxis1 = image_header1['naxis1']
    naxis2 = image_header1['naxis2']

    # if required, display file1
    if display == 'all':
        ximshow_file(file1,
                     extnum=extnum_from_extname(file1, extname1)+1,  # extnum is 1-based in ximshow_file
                     args_z1z2=args_z1z2, args_bbox=args_bbox,
                     args_keystitle=args_keystitle,
                     args_geometry=args_geometry,
                     debugplot=12)

    # read second FITS file or number
    file2_is_a_file = True
    try:
        with fits.open(file2) as hdulist:
            image_header2 = hdulist[extname2].header
            image2 = hdulist[extname2].data.astype(dtype)
            naxis_ = image_header2['naxis']
            naxis1_ = image_header2['naxis1']
            naxis2_ = image_header2['naxis2']
            filename = file2
    except FileNotFoundError:
        file2_is_a_file = False
        image2 = np.zeros((naxis2, naxis1), dtype=dtype)
        if 'int' in dtype:
            image2 += int(file2)
        elif 'float' in dtype:
            image2 += float(file2)
        else:
            raise ValueError(f"Unsupported dtype: {dtype}. Use an integer or float type.")
        naxis_ = naxis
        naxis1_ = naxis1
        naxis2_ = naxis2
        filename = 'constant=' + str(file2)

    # if required, display file2
    if display == 'all':
        if file2_is_a_file:
            ximshow_file(file2,
                         extnum=extnum_from_extname(file2, extname2)+1,  # extnum is 1-based in ximshow_file
                         args_z1z2=args_z1z2, args_bbox=args_bbox,
                         args_keystitle=args_keystitle,
                         args_geometry=args_geometry,
                         debugplot=12)
        else:
            ximshow(image2,
                    title=filename,
                    z1z2=args_z1z2, image_bbox=args_bbox,
                    geometry=args_geometry,
                    debugplot=12)

    # second image is a single image row
    if naxis1 == naxis1_ and naxis2 > naxis2_:
        if naxis2_ == 1:
            image2 = np.tile(image2, naxis2).reshape((naxis2, naxis1))
            naxis2_ = naxis2

    # second image is a single image column
    if naxis2 == naxis2_ and naxis1 > naxis1_:
        if naxis1_ == 1:
            image2 = np.repeat(image2, naxis1).reshape((naxis2, naxis1))
            naxis1_ = naxis1

    # additional dimension checks
    if naxis != naxis_:
        raise ValueError("NAXIS values are different.")
    if naxis1 != naxis1_:
        raise ValueError("NAXIS1 values are different.")
    if naxis2 != naxis2_:
        raise ValueError("NAXIS2 values are different.")

    # compute operation
    if operation == "+":
        solution = image1 + image2
    elif operation == "-":
        solution = image1 - image2
    elif operation == "x":
        solution = image1 * image2
    elif operation == "/":
        solution = image1 / image2
    elif operation == "=":
        solution = image2.copy()
    else:
        raise ValueError("Unexpected operation=" + str(operation))

    # save output file
    hdu = fits.PrimaryHDU(solution.astype(dtype), image_header1)
    hdu.writeto(output, overwrite=overwrite)

    # if required, display result
    if display in ['all', 'result']:
        ximshow_file(output,
                     args_z1z2=args_z1z2, args_bbox=args_bbox,
                     args_keystitle=args_keystitle,
                     args_geometry=args_geometry,
                     debugplot=12)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: binary image arithmetic"
    )
    # positional parameters
    parser.add_argument("file1",
                        help="First FITS image",
                        type=argparse.FileType('rb'))
    parser.add_argument("operation",
                        help="Arithmetic operation",
                        type=str,
                        choices=['+', '-', 'x', '/', '='])
    parser.add_argument("file2",
                        help="Second FITS image or number",
                        type=str)
    parser.add_argument("output",
                        help="Output FITS image",
                        type=str)
    # optional arguments
    parser.add_argument("--extname1", type=str,
                        help="Extension name of the first FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--extname2", type=str,
                        help="Extension name of the second FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--overwrite",
                        help="Overwrite output file if already exists",
                        action="store_true")
    parser.add_argument("--dtype", 
                        help="Data type of the output image (default: float32)",
                        type=str, 
                        choices=['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64',
                                 'float32', 'float64'],
                        default='float32')
    parser.add_argument("--display",
                        help="Display images: all, result, none (default)",
                        default="none",
                        type=str,
                        choices=['all', 'result', 'none'])
    parser.add_argument("--z1z2",
                        help="tuple z1,z2, minmax or None (use zscale)")
    parser.add_argument("--bbox",
                        help="bounding box tuple: nc1,nc2,ns1,ns2")
    parser.add_argument("--keystitle",
                        help="tuple of FITS keywords.format: " +
                             "key1,key2,...keyn.'format'")
    parser.add_argument("--geometry",
                        help="Tuple x,y,dx,dy indicating window geometry",
                        default="0,0,640,480")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)


    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if not args.overwrite and pathlib.Path(args.output).exists():
        print(f"Output file {args.output} already exists. Use --overwrite to overwrite it.")
        sys.exit(1)

    # compute operation
    compute_operation(
        file1=args.file1,
        file2=args.file2,
        operation=args.operation,
        extname1=args.extname1,
        extname2=args.extname2,
        output=args.output,
        overwrite=args.overwrite,
        dtype=args.dtype,
        display=args.display,
        args_z1z2=args.z1z2,
        args_bbox=args.bbox,
        args_keystitle=args.keystitle,
        args_geometry=args.geometry
    )


if __name__ == "__main__":

    main()



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Quality control for Numina-based applications."""

import enum


class QC(enum.Enum):
    GOOD = 1
    PARTIAL = 2
    BAD = 3
    UNKNOWN = 4



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Generate a 3D mosaic from individual 3D cubes.
"""

import argparse
from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from datetime import datetime
import numpy as np
from pathlib import Path
from reproject import reproject_interp, reproject_adaptive, reproject_exact
from reproject.mosaicking import find_optimal_celestial_wcs
import sys

from numina.array.array_size_32bits import array_size_8bits, array_size_32bits
from numina.instrument.simulation.ifu.define_3d_wcs import header3d_after_merging_wcs2d_celestial_and_wcs1d_spectral
from numina.instrument.simulation.ifu.define_3d_wcs import wcs_to_header_using_cd_keywords

from .add_script_info_to_fits_history import add_script_info_to_fits_history
from .ctext import ctext
from .resample_wave_3d_cube import resample_wave_3d_cube

REPROJECT_METHODS = ['interp', 'adaptive', 'exact']


def generate_mosaic_of_3d_cubes(
        list_of_fits_files,
        extname_image,
        crval3out,
        cdelt3out,
        naxis3out,
        desired_celestial_2d_wcs,
        reproject_method,
        parallel,
        output_celestial_2d_wcs,
        footprint=False,
        verbose=False
):
    """Combine 3D cubes using their WCS information.

    Parameters
    ----------
    list_of_fits_files : list of str
        List of paths to FITS files containing the 3D cubes to be combined.
    extname_image : str
        Extension name for the image in input files.
    crval3out : `astropy.units.Quantity` or None
        Minimum wavelength (in m) for the output image. If None,
        use the minimum wavelength of the input images.
    cdelt3out : `astropy.units.Quantity` or None
        Wavelength step (in m/pixel) for the output image. If None,
        the output image will use the minimum wavelength step of the input images.
    naxis3out : int or None
        Number of slices in the output image. If None, the output image
        will have the required number of slices to cover the full wavelength
        range of the input images.
    desired_celestial_2d_wcs : str, file-like, `pathlib.Path` or None
        FITS filename with desired 2D celestial WCS. If None,
        compute output 2D celestial WCS for current list of 3D cubes.
    reproject_method : str
        Reprojection method. See 'REPROJECT_METHODS' above.
    parallel : bool
        If True, use parallel processing for reprojection.
    output_celestial_2d_wcs : str, file-like, `pathlib.Path` or None
        Path to output 2D celestial WCS.
    footprint : bool
        If True, generate a FOOTPRINT extension with the final footprint.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    output_hdul : `astropy.io.fits.HDUList`
        Instance of HDUList with two HDU:
        - PRIMARY: mosaic image
        - FOOTPRINT: array with final footprint
    """
    # protections
    if not isinstance(list_of_fits_files, list):
        raise TypeError('list_of_fits_files must be a list')
    if crval3out is not None and not isinstance(crval3out, u.Quantity):
        raise TypeError('crval3out must be an astropy Quantity')
    if cdelt3out is not None and not isinstance(cdelt3out, u.Quantity):
        raise TypeError('cdelt3out must be an astropy Quantity')
    if naxis3out is not None and not isinstance(naxis3out, int):
        raise TypeError('naxis3out must be an integer')

    nimages = len(list_of_fits_files)
    if verbose:
        print(f'Total number of images to be combined: {nimages}')

    if nimages < 1:
        raise ValueError('Number of images = 0')

    # compute crval3out, cdelt3out and naxis3out if not provided
    crval3out_ = None
    wavemax = None   # maximum wavelength (at the center of the last pixel)
    cdelt3out_ = None
    for fname in list_of_fits_files:
        with fits.open(fname) as hdul:
            if extname_image not in hdul:
                raise ValueError(f'Expected {extname_image} extension not found')
            hdu = hdul[extname_image]
            if verbose:
                print(f'{hdu.header["NAXIS1"]=}, {hdu.header["NAXIS2"]=}, {hdu.header["NAXIS3"]=}')
            wcs1d_spectral = WCS(hdu.header).spectral
            wave = wcs1d_spectral.pixel_to_world(np.arange(hdu.data.shape[0]))
        if crval3out_ is None:
            crval3out_ = wave[0]
        else:
            if wave[0] < crval3out_:
                crval3out_ = wave[0]
        if wavemax is None:
            wavemax = wave[-1]
        else:
            if wave[-1] > wavemax:
                wavemax = wave[-1]
        if cdelt3out_ is None:
            cdelt3out_ = np.diff(wave).min()
        else:
            if np.diff(wave).min() < cdelt3out_:
                cdelt3out_ = np.diff(wave).min()
    if crval3out is None:
        crval3out = crval3out_
    if cdelt3out is None:
        cdelt3out = cdelt3out_ / u.pix
    if naxis3out is None:
        naxis3out = int(np.round((wavemax.value - crval3out.value) / cdelt3out.value)) + 1
    wavemax = crval3out + cdelt3out * (naxis3out - 1) * u.pix
    # define 1D WCS for the spectral axis of the combined mosaic
    header_spectral_mosaic = fits.Header()
    header_spectral_mosaic['NAXIS'] = 1
    header_spectral_mosaic['NAXIS1'] = naxis3out
    header_spectral_mosaic['CRPIX1'] = 1.0
    header_spectral_mosaic['CDELT1'] = cdelt3out.to(u.m/u.pix).value
    header_spectral_mosaic['CRVAL1'] = crval3out.to(u.m).value
    header_spectral_mosaic['CUNIT1'] = 'm'
    header_spectral_mosaic['CTYPE1'] = 'WAVE'
    wcs1d_spectral_mosaic = WCS(header_spectral_mosaic)
    if verbose:
        print(f'\n{crval3out=}\n{cdelt3out=}\n{naxis3out=}\n{wavemax=}\n')
        print(f'{wcs1d_spectral_mosaic=}')

    # optimal 2D WCS (celestial part) for combined mosaic
    if desired_celestial_2d_wcs is None:
        if verbose:
            print('\nCelestial scales:')
        # compute final celestial WCS for the ensemble of 3D cubes
        list_of_inputs = []
        for i, fname in enumerate(list_of_fits_files):
            with fits.open(fname) as hdul:
                hdu = hdul[extname_image]
            header3d = hdu.header
            wcs2d = WCS(header3d).celestial
            scales = proj_plane_pixel_scales(wcs2d)
            if verbose:
                print(f'Image {i+1}: {scales[0]*3600:.3f} arcsec, {scales[1]*3600:.3f} arcsec')
            list_of_inputs.append(((header3d['NAXIS2'], header3d['NAXIS1']), wcs2d))
        wcs_mosaic2d, shape_mosaic2d = find_optimal_celestial_wcs(list_of_inputs)
        scales = proj_plane_pixel_scales(wcs_mosaic2d)
        if verbose:
            print(f'Mosaic : {scales[0]*3600:.3f} arcsec, {scales[1]*3600:.3f} arcsec')
    else:
        # make use of an external celestial WCS projection
        if verbose:
            print(ctext(f'\nUsing external celestial WCS: {desired_celestial_2d_wcs}', fg='green'))
        with fits.open(desired_celestial_2d_wcs) as hdul_mosaic2d:
            wcs_mosaic2d = WCS(hdul_mosaic2d[0].header)
            shape_mosaic2d = hdul_mosaic2d[0].header['NAXIS2'], hdul_mosaic2d[0].header['NAXIS1']
    if verbose:
        print(f'\n{wcs_mosaic2d=}')
        print(f'\n{shape_mosaic2d=}')
    if output_celestial_2d_wcs is not None:
        header_2d_wcs = wcs_to_header_using_cd_keywords(wcs_mosaic2d)
        hdu = fits.PrimaryHDU(np.zeros(shape_mosaic2d, dtype=np.uint8), header=header_2d_wcs)
        if verbose:
            print(f'Saving resulting celestial 2D WCS to: {output_celestial_2d_wcs}')
        hdu.writeto(output_celestial_2d_wcs, overwrite=True)

    # initialize arrays to store combination
    naxis3_mosaic3d = naxis3out
    naxis2_mosaic3d, naxis1_mosaic3d = shape_mosaic2d
    mosaic3d_cube_by_cube = np.zeros((naxis3_mosaic3d, naxis2_mosaic3d, naxis1_mosaic3d))
    footprint3d = np.zeros(shape=(naxis3_mosaic3d, naxis2_mosaic3d, naxis1_mosaic3d))
    if verbose:
        print(f'\nNAXIS1, NAXIS2, NAXIS3 of 3D mosaic: {naxis1_mosaic3d}, {naxis2_mosaic3d}, {naxis3_mosaic3d}')
        size_output = array_size_32bits(mosaic3d_cube_by_cube)
        if footprint:
            size_output += array_size_8bits(footprint3d)
        print(ctext(f'Combined image will require {size_output:.2f}', fg='red'))

    # generate 3D mosaic
    for fname in list_of_fits_files:
        time_ini = datetime.now()
        if verbose:
            print(f'\n* Working with: {fname}')
        with fits.open(fname) as hdul:
            hdu = hdul[extname_image]
            single_hdu3d = resample_wave_3d_cube(
                hdu3d_image=hdu,
                crval3out=crval3out,
                cdelt3out=cdelt3out,
                naxis3out=naxis3out,
                verbose=verbose
            )
        data_ini3d = single_hdu3d.data
        wcs_ini3d = WCS(single_hdu3d.header)
        wcs_ini2d = wcs_ini3d.celestial
        if verbose:
            print(f'Celestial WCS reprojection method: {reproject_method}')
        if reproject_method == 'interp':
            temp3d, footprint_temp3d = reproject_interp(
                (data_ini3d, wcs_ini2d),
                wcs_mosaic2d,
                shape_out=shape_mosaic2d,
                parallel=parallel
            )
        elif reproject_method == 'adaptive':
            temp3d, footprint_temp3d = reproject_adaptive(
                (data_ini3d, wcs_ini2d),
                wcs_mosaic2d,
                shape_out=shape_mosaic2d,
                conserve_flux=True,
                kernel='Gaussian',
                parallel=parallel
            )
        elif reproject_method == 'exact':
            temp3d, footprint_temp3d = reproject_exact(
                (data_ini3d, wcs_ini2d),
                wcs_mosaic2d,
                shape_out=shape_mosaic2d,
                parallel=parallel
            )
        else:
            raise ValueError(f'Unexpected {reproject_method=}')
        valid_region = (footprint_temp3d > 0)
        mosaic3d_cube_by_cube[valid_region] += temp3d[valid_region]
        footprint3d += footprint_temp3d
        time_end = datetime.now()
        if verbose:
            print(f'Processing time for {fname}: {time_end - time_ini}')

    valid_region = (footprint3d > 0)
    mosaic3d_cube_by_cube[valid_region] /= footprint3d[valid_region]
    invalid_region = (footprint3d == 0)
    mosaic3d_cube_by_cube[invalid_region] = np.nan  # set invalid pixels to NaN

    # generate result
    hdu = fits.PrimaryHDU(mosaic3d_cube_by_cube.astype(np.float32))
    header3d_corrected = header3d_after_merging_wcs2d_celestial_and_wcs1d_spectral(
        wcs2d_celestial=wcs_mosaic2d,
        wcs1d_spectral=wcs1d_spectral_mosaic
    )
    hdu.header.update(header3d_corrected)
    if footprint:
        hdu_footprint = fits.ImageHDU(footprint3d.astype(np.uint8))
        hdu_footprint.header['EXTNAME'] = 'FOOTPRINT'
        hdu_footprint.header.update(header3d_corrected)
        output_hdul = fits.HDUList([hdu, hdu_footprint])
    else:
        output_hdul = fits.HDUList([hdu])

    return output_hdul


def main(args=None):

    time_ini = datetime.now()
    # parse command-line options
    parser = argparse.ArgumentParser(
        description="Generate a 3D mosaic from individual 3D cubes."
    )
    parser.add_argument("input_list",
                        help="TXT file with list of 3D images to be combined or single FITS file", type=str)
    parser.add_argument('output_filename',
                        help='filename of output FITS image', type=str)
    parser.add_argument("--crval3out", help="Minimum wavelength (in m) for the output image",
                        type=float, default=None)
    parser.add_argument("--cdelt3out", help="Wavelength step (in m/pixel) for the output image",
                        type=float, default=None)
    parser.add_argument("--naxis3out", help="Number of slices in the output image", type=int, default=None)
    parser.add_argument("--desired_celestial_2d_wcs",
                        help="Desired 2D celestial WCS projection. "
                        "Default None (compute for current 3D cube combination)",
                        type=str, default=None)
    parser.add_argument('--reproject_method',
                        help='Reprojection method (interp, adaptive, exact)',
                        type=str, choices=REPROJECT_METHODS, default='adaptive')
    parser.add_argument('--parallel',
                        help='Use parallel processing for reprojection',
                        action='store_true')
    parser.add_argument('--extname_image',
                        help='Extension name for image in input files. Default value: PRIMARY',
                        default='PRIMARY', type=str)
    parser.add_argument("--output_celestial_2d_wcs",
                        help="filename for output 2D celestial WCS",
                        type=str, default=None)
    parser.add_argument("--footprint",
                        help="Generate a FOOTPRINT extension with the final footprint",
                        action="store_true")
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_list = args.input_list
    output_filename = args.output_filename
    crval3out = args.crval3out
    if crval3out is not None:
        crval3out = crval3out * u.m
    cdelt3out = args.cdelt3out
    if cdelt3out is not None:
        cdelt3out = cdelt3out * u.m / u.pix
    naxis3out = args.naxis3out
    desired_celestial_2d_wcs = args.desired_celestial_2d_wcs
    reproject_method = args.reproject_method
    parallel = args.parallel
    if reproject_method not in REPROJECT_METHODS:
        raise ValueError(f'Unexpected reproject_method: {reproject_method}. Expected one of {REPROJECT_METHODS}')
    output_celestial_2d_wcs = args.output_celestial_2d_wcs
    footprint = args.footprint
    verbose = args.verbose

    # define extensions for image and mask
    extname_image = args.extname_image

    # check if input file is a single FITS file or a list
    if input_list.endswith('.fits'):
        file_content = [input_list]
    else:
        with open(input_list) as f:
            file_content = f.read().splitlines()

    list_of_fits_files = []
    for fname in file_content:
        if len(fname) > 0:
            if fname[0] not in ['#']:
                if not Path(fname).is_file():
                    raise ValueError(f'File {fname} does not exist or is not a valid file.')
                list_of_fits_files.append(fname)

    if len(list_of_fits_files) < 1:
        raise ValueError(f'No valid FITS files found in {input_list}. Please check the file content.')

    # combine images
    output_hdul = generate_mosaic_of_3d_cubes(
        list_of_fits_files=list_of_fits_files,
        extname_image=extname_image,
        crval3out=crval3out,
        cdelt3out=cdelt3out,
        naxis3out=naxis3out,
        desired_celestial_2d_wcs=desired_celestial_2d_wcs,
        reproject_method=reproject_method,
        parallel=parallel,
        output_celestial_2d_wcs=output_celestial_2d_wcs,
        footprint=footprint,
        verbose=verbose
    )

    # save result
    add_script_info_to_fits_history(output_hdul[0].header, args)
    output_hdul[0].header.add_history('Contents of --input_list:')
    for item in list_of_fits_files:
        output_hdul[0].header.add_history(f'- {item}')
    if verbose:
        print(f'Saving: {output_filename}')
    output_hdul.writeto(output_filename, overwrite='yes')

    time_end = datetime.now()
    if verbose:
        print(f'\nTotal time: {time_end - time_ini}')
        print('Done!')


if __name__ == "__main__":
    main()



import sys

import pytest
import numpy
import astropy.io.fits as fits

from ..correction import header_add_barycentric_correction


def create_header1():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2017-08-23T21:38:30.55'
    # GTC
    hdr['OBSGEO-X'] = 5327285.0921
    hdr['OBSGEO-Y'] = -1718777.1125
    hdr['OBSGEO-Z'] = 3051786.7327

    hdr['RADEG'] = 285.481037748898
    hdr['DECDEG'] = 42.4882140636786

    hdr['CTYPE1'] = 'AWAV'
    hdr['CRPIX1'] = 1
    hdr['CRVAL1'] = 362.0
    hdr['CDELT1'] = 1.86
    hdr['CUNIT1'] = 'nm'

    hdr['CRPIX2'] = 0
    hdr['CRVAL2'] = 0
    hdr['CDELT2'] = 1
    return hdr


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires python3.5 or higher")
def test_add_barycentric():
    import astropy.wcs
    import astropy.time
    from astropy.coordinates import SkyCoord, EarthLocation
    import astropy.constants as cons

    hdr = create_header1()
    hdr = header_add_barycentric_correction(hdr, key='b')

    # Get main WCS
    wcs0 = astropy.wcs.WCS(hdr)

    gtc = EarthLocation.from_geocentric(
        wcs0.wcs.obsgeo[0], wcs0.wcs.obsgeo[1], wcs0.wcs.obsgeo[2], unit='m')
    date_obs = astropy.time.Time(wcs0.wcs.dateobs, format='fits')
    # if frame='fk5', we need to pass the epoch and equinox
    sc = SkyCoord(ra=hdr['RADEG'], dec=hdr['DECDEG'], unit='deg')
    rv = sc.radial_velocity_correction(obstime=date_obs, location=gtc)
    factor = (1 + rv / cons.c).to('').value

    # velocity
    rv = rv.to('m / s').value
    # (1 + rv / c)
    # factor = 0.9999762685198925

    assert hdr['WCSNAMEB'] == 'Barycentric correction'
    assert hdr['CTYPE1B'] == hdr['CTYPE1']
    assert hdr['CRPIX1B'] == hdr['CRPIX1']
    assert numpy.allclose(hdr['CRVAL1B'], hdr['CRVAL1'] * factor)
    assert numpy.allclose(hdr['CDELT1B'], hdr['CDELT1'] * factor)
    assert hdr['CUNIT1B'] == hdr['CUNIT1']
    #
    assert hdr['CRPIX2B'] == hdr['CRPIX2']
    assert hdr['CRVAL2B'] == hdr['CRVAL2']
    assert hdr['CDELT2B'] == hdr['CDELT2']
    #
    assert numpy.allclose(hdr['VELOSYSB'], rv)
    assert hdr['SPECSYSB'] == 'BARYCENT'
    assert hdr['SSYSOBSB'] == 'TOPOCENT'


def test_add_barycentric_missing1():
    hdr = create_header1()
    del hdr['RADEG']

    with pytest.raises(KeyError):
        header_add_barycentric_correction(hdr, key='b')


def test_add_barycentric_missing2():
    hdr = create_header1()
    del hdr['DATE-OBS']

    with pytest.raises(KeyError):
        header_add_barycentric_correction(hdr, key='b')


def test_add_barycentric_missing3():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2017-08-23T21:38:30.55'
    # GTC
    hdr['OBSGEO-X'] = 5327285.0921
    hdr['OBSGEO-Y'] = -1718777.1125
    hdr['OBSGEO-Z'] = 3051786.7327

    hdr['RADEG'] = 285.481037748898
    hdr['DECDEG'] = 42.4882140636786

    with pytest.raises(TypeError):
        header_add_barycentric_correction(hdr, key='b')



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from scipy.stats import norm
import numpy as np
from astropy.modeling import Fittable1DModel, Parameter
import math

M_SQRT_2_PI = math.sqrt(2 * math.pi)


def norm_pdf_t(x):
    return np.exp(-0.5 * x * x) / M_SQRT_2_PI


def gauss_box_model(x, amplitude=1.0, mean=0.0, stddev=1.0, hpix=0.5):
    """Integrate a Gaussian profile."""
    z = (x - mean) / stddev
    z2 = z + hpix / stddev
    z1 = z - hpix / stddev
    return amplitude * (norm.cdf(z2) - norm.cdf(z1))


def gauss_box_model_deriv(x, amplitude=1.0, mean=0.0, stddev=1.0, hpix=0.5):
    """Derivative of the integral of  a Gaussian profile."""
    z = (x - mean) / stddev
    z2 = z + hpix / stddev
    z1 = z - hpix / stddev

    da = norm.cdf(z2) - norm.cdf(z1)

    fp2 = norm_pdf_t(z2)
    fp1 = norm_pdf_t(z1)

    dl = -amplitude / stddev * (fp2 - fp1)
    ds = -amplitude / stddev * (fp2 * z2 - fp1 * z1)
    dd = amplitude / stddev * (fp2 + fp1)

    return da, dl, ds, dd


class GaussBox(Fittable1DModel):
    """Model for fitting a 1D Gaussina convolved with a square"""
    amplitude = Parameter(default=1.0)
    mean = Parameter(default=0.0)
    stddev = Parameter(default=1.0)
    hpix = Parameter(default=0.5, fixed=True)

    @staticmethod
    def evaluate(x, amplitude, mean, stddev, hpix):
        return gauss_box_model(x, amplitude, mean, stddev, hpix)

    @staticmethod
    def fit_deriv(x, amplitude, mean, stddev, hpix):
        return gauss_box_model_deriv(x, amplitude, mean, stddev, hpix)



from ..assembly import process_setup
from ..elements import SetupBlock


def test_setup1():
    setup_block = [{"values": {"ypos": [1, 2, 3, 4, 5, 6, 7]}}]
    res = process_setup("comp_store", "test", setup_block, "date")
    assert isinstance(res, SetupBlock)
